from __future__ import unicode_literals
from __future__ import print_function

from threading import Thread, RLock, Event
from datetime import datetime, timedelta

from operator import attrgetter


import logging
log = logging.getLogger('moya.runtime')


class Task(object):
    """A task scheduled for a given time"""
    def __init__(self, name, callable, run_time=None, repeat=None):
        self.name = name
        self.callable = callable
        self.run_time = run_time
        self.repeat = repeat

    def __unicode__(self):
        if self.repeat:
            return "<task '{}' repeats {}>".format(self.name, self.repeat)
        else:
            return "<task '{}'>".format(self.name)

    __repr__ = __unicode__

    def advance(self):
        """Set run time to next cycle"""
        if self.repeat:
            self.run_time = self.run_time + self.repeat
            return True
        else:
            return False

    def ready(self, t):
        """Check if this task is ready to be execute"""
        return t >= self.run_time

    def run(self):
        """Run the task in it's own thread"""
        run_thread = Thread(target=self._run)
        run_thread.start()

    def _run(self):
        try:
            self.callable()
        except Exception as e:
            log.exception("Error in scheduled task '%s'", self.name)


class Scheduler(Thread):
    """A thread that manages scheduled / repeating tasks"""

    # The granularity at which tasks can be scheduled
    # Super precision is not required given that tasks are unlikely to be more than one a minute
    poll_seconds = 1

    def __init__(self):
        super(Scheduler, self).__init__()
        self.lock = RLock()
        self.quit_event = Event()
        self.tasks = []
        self.new_tasks = []

    def run(self):
        while 1:
            # Better than a sleep because we can interrupt it
            self.quit_event.wait(self.poll_seconds)
            if self.quit_event.is_set():
                break
            with self.lock:
                self.tasks += self.new_tasks
                del self.new_tasks[:]
            self._check_tasks()
        del self.tasks[:]
        with self.lock:
            del self.new_tasks[:]

    def stop(self, wait=True):
        """Stop processing tasks and block till everything has quit"""
        self.quit_event.set()
        if wait:
            self.join()

    def _check_tasks(self):
        if not self.tasks:
            return
        t = datetime.utcnow()
        ready_tasks = None
        self.tasks.sort(key=attrgetter('run_time'))
        for i, task in enumerate(self.tasks):
            if not task.ready(t):
                ready_tasks = self.tasks[:i]
                del self.tasks[:i]
                break
        else:
            ready_tasks = self.tasks[:]
            del self.tasks[:]

        if ready_tasks:
            for task in ready_tasks:
                if self.quit_event.is_set():
                    break
                task.run()
                if task.advance():
                    self.tasks.append(task)

    def add_repeat_task(self, name, callable, repeat):
        """Add a task to be invoked every `repeat` seconds"""
        if isinstance(repeat, (int, long, float)):
            repeat = timedelta(seconds=repeat)
        run_time = datetime.utcnow() + repeat
        task = Task(name, callable, run_time=run_time, repeat=repeat)
        with self.lock:
            self.new_tasks.append(task)
        return task


if __name__ == "__main__":
    import time
    def hello():
        print("Hello")
    def world():
        print("world")
    scheduler = Scheduler()
    print(scheduler.add_repeat_task("hello", hello, repeat=5))
    print(scheduler.add_repeat_task("world", world, repeat=3))
    scheduler.start()

    try:
        while 1:
            time.sleep(1)
    except:
        scheduler.stop()
        print("Stopped")
        raise



import unittest
import os

from fs.opener import fsopendir

from moya.context import Context
from moya.console import Console
from moya.archive import Archive


class TestCalls(unittest.TestCase):

    def setUp(self):
        self.called = False
        path = os.path.abspath(os.path.dirname(__file__))
        self.fs = fsopendir(path)
        self.context = Context()
        self.context['console'] = Console()
        self.archive = Archive()
        import_fs = self.fs.opendir("archivetest")
        self.archive.load_library(import_fs)
        self.archive.finalize()

    def test_moya_call_no_lazy(self):
        """Test moya call without lazy attribute"""
        self.archive('moya.tests#test_moya_call_no_lazy', self.context, None)
        self.assert_(self.context.root['called'])
        self.assertEqual(self.context['.result'], 123)

    def test_moya_call_lazy(self):
        """Test lazy moya calls"""
        self.archive('moya.tests#test_moya_call_lazy', self.context, None)
        self.assert_('called' not in self.context.root)
        self.assertEqual(self.context['result'], 123)
        self.assert_('called' in self.context.root)
        self.assert_(self.context.root['called'])

    # def test_call_no_lazy(self):
    #     """Test callable call without lazy attribute"""
    #     def callable(a, b):
    #         return a + b
    #     self.archive('moya.tests#test_call_no_lazy',
    #                  self.context,
    #                  callable=callable,
    #                  a=100,
    #                  b=23)
    #     self.assertEqual(self.context['.result'], 123)

    # def test_call_lazy(self):
    #     """Test callable call without lazy attribute"""
    #     def callable(a, b):
    #         self.called = True
    #         return a + b
    #     # Sanity check that there has been no call
    #     self.assert_('result' not in self.context)
    #     # Call callable lazily store the result in context
    #     self.archive('moya.tests#test_call_lazy',
    #                  self.context,
    #                  callable=callable,
    #                  a=100,
    #                  b=23)
    #     # Check callble has not been called yet
    #     self.assert_(not self.called)
    #     # Check lazy result is in context
    #     self.assert_('result' in self.context)
    #     # Check lazy callable still hasn't been called
    #     self.assert_(not self.called)
    #     # Check return value (should call callable)
    #     self.assertEqual(self.context['.result'], 123)
    #     # Check callable has been called
    #     self.assert_(self.called)
    #     # Check cached result
    #     self.assertEqual(self.context['.result'], 123)



"""

A color context object

"""


from __future__ import unicode_literals
from __future__ import division
from __future__ import print_function

from ..compat import text_type, implements_to_string
from ..interface import AttributeExposer

from collections import OrderedDict
import colorsys
import re

# Borrowed from https://github.com/bahamas10/css-color-names/blob/master/css-color-names.json
HTML_COLORS = {
    "aqua": "#00ffff",
    "aliceblue": "#f0f8ff",
    "antiquewhite": "#faebd7",
    "black": "#000000",
    "blue": "#0000ff",
    "cyan": "#00ffff",
    "darkblue": "#00008b",
    "darkcyan": "#008b8b",
    "darkgreen": "#006400",
    "darkturquoise": "#00ced1",
    "deepskyblue": "#00bfff",
    "green": "#008000",
    "lime": "#00ff00",
    "mediumblue": "#0000cd",
    "mediumspringgreen": "#00fa9a",
    "navy": "#000080",
    "springgreen": "#00ff7f",
    "teal": "#008080",
    "midnightblue": "#191970",
    "dodgerblue": "#1e90ff",
    "lightseagreen": "#20b2aa",
    "forestgreen": "#228b22",
    "seagreen": "#2e8b57",
    "darkslategray": "#2f4f4f",
    "darkslategrey": "#2f4f4f",
    "limegreen": "#32cd32",
    "mediumseagreen": "#3cb371",
    "turquoise": "#40e0d0",
    "royalblue": "#4169e1",
    "steelblue": "#4682b4",
    "darkslateblue": "#483d8b",
    "mediumturquoise": "#48d1cc",
    "indigo": "#4b0082",
    "darkolivegreen": "#556b2f",
    "cadetblue": "#5f9ea0",
    "cornflowerblue": "#6495ed",
    "mediumaquamarine": "#66cdaa",
    "dimgray": "#696969",
    "dimgrey": "#696969",
    "slateblue": "#6a5acd",
    "olivedrab": "#6b8e23",
    "slategray": "#708090",
    "slategrey": "#708090",
    "lightslategray": "#778899",
    "lightslategrey": "#778899",
    "mediumslateblue": "#7b68ee",
    "lawngreen": "#7cfc00",
    "aquamarine": "#7fffd4",
    "chartreuse": "#7fff00",
    "gray": "#808080",
    "grey": "#808080",
    "maroon": "#800000",
    "olive": "#808000",
    "purple": "#800080",
    "lightskyblue": "#87cefa",
    "skyblue": "#87ceeb",
    "blueviolet": "#8a2be2",
    "darkmagenta": "#8b008b",
    "darkred": "#8b0000",
    "saddlebrown": "#8b4513",
    "darkseagreen": "#8fbc8f",
    "lightgreen": "#90ee90",
    "mediumpurple": "#9370db",
    "darkviolet": "#9400d3",
    "palegreen": "#98fb98",
    "darkorchid": "#9932cc",
    "yellowgreen": "#9acd32",
    "sienna": "#a0522d",
    "brown": "#a52a2a",
    "darkgray": "#a9a9a9",
    "darkgrey": "#a9a9a9",
    "greenyellow": "#adff2f",
    "lightblue": "#add8e6",
    "paleturquoise": "#afeeee",
    "lightsteelblue": "#b0c4de",
    "powderblue": "#b0e0e6",
    "firebrick": "#b22222",
    "darkgoldenrod": "#b8860b",
    "mediumorchid": "#ba55d3",
    "rosybrown": "#bc8f8f",
    "darkkhaki": "#bdb76b",
    "silver": "#c0c0c0",
    "mediumvioletred": "#c71585",
    "indianred": "#cd5c5c",
    "peru": "#cd853f",
    "chocolate": "#d2691e",
    "tan": "#d2b48c",
    "lightgray": "#d3d3d3",
    "lightgrey": "#d3d3d3",
    "thistle": "#d8bfd8",
    "goldenrod": "#daa520",
    "orchid": "#da70d6",
    "palevioletred": "#db7093",
    "crimson": "#dc143c",
    "gainsboro": "#dcdcdc",
    "plum": "#dda0dd",
    "burlywood": "#deb887",
    "lightcyan": "#e0ffff",
    "lavender": "#e6e6fa",
    "darksalmon": "#e9967a",
    "palegoldenrod": "#eee8aa",
    "violet": "#ee82ee",
    "azure": "#f0ffff",
    "honeydew": "#f0fff0",
    "khaki": "#f0e68c",
    "lightcoral": "#f08080",
    "sandybrown": "#f4a460",
    "beige": "#f5f5dc",
    "mintcream": "#f5fffa",
    "wheat": "#f5deb3",
    "whitesmoke": "#f5f5f5",
    "ghostwhite": "#f8f8ff",
    "lightgoldenrodyellow": "#fafad2",
    "linen": "#faf0e6",
    "salmon": "#fa8072",
    "oldlace": "#fdf5e6",
    "bisque": "#ffe4c4",
    "blanchedalmond": "#ffebcd",
    "coral": "#ff7f50",
    "cornsilk": "#fff8dc",
    "darkorange": "#ff8c00",
    "deeppink": "#ff1493",
    "floralwhite": "#fffaf0",
    "fuchsia": "#ff00ff",
    "gold": "#ffd700",
    "hotpink": "#ff69b4",
    "ivory": "#fffff0",
    "lavenderblush": "#fff0f5",
    "lemonchiffon": "#fffacd",
    "lightpink": "#ffb6c1",
    "lightsalmon": "#ffa07a",
    "lightyellow": "#ffffe0",
    "magenta": "#ff00ff",
    "mistyrose": "#ffe4e1",
    "moccasin": "#ffe4b5",
    "navajowhite": "#ffdead",
    "orange": "#ffa500",
    "orangered": "#ff4500",
    "papayawhip": "#ffefd5",
    "peachpuff": "#ffdab9",
    "pink": "#ffc0cb",
    "red": "#ff0000",
    "seashell": "#fff5ee",
    "snow": "#fffafa",
    "tomato": "#ff6347",
    "white": "#ffffff",
    "yellow": "#ffff00",
    "rebeccapurple": "#663399"
}

# Add 'special' colors
HTML_COLORS['transparent'] = "rgba(0,0,0,0)"


@implements_to_string
class Color(AttributeExposer):
    """HTML color object"""

    __moya_exposed_attributes__ = [
        "r", "g", "b", "a",
        "rgb", "rgba", "hsl", "hsla",
        "html", "hex",
        "name",
        "opaque", "saturated"
    ]

    _re_hex = re.compile(r'^\#([a-fA-F0-9]{2})([a-fA-F0-9]{2})([a-fA-F0-9]{2})$')
    _re_hex_short = re.compile(r'^\#([a-fA-F0-9]{1})([a-fA-F0-9]{1})([a-fA-F0-9]{1})$')
    _re_rgb = re.compile(r'rgb\(\s*([0-9\.]+)\s*,\s*([0-9\.]+)\s*,\s*([0-9\.]+)\s*\)')
    _re_rgba = re.compile(r'rgba\(\s*([0-9\.]+)\s*,\s*([0-9\.]+)\s*,\s*([0-9\.]+)\s*,\s*([0-9\.]+)\s*\)')
    _re_hsl = re.compile(r'hsl\(\s*([0-9\.]+)\s*,\s*([0-9\.]+)\%\s*,\s*([0-9\.]+)\%\s*\)')
    _re_hsla = re.compile(r'hsla\(\s*([0-9\.]+)\s*,\s*([0-9\.]+)\%\s*,\s*([0-9\.]+)\%\s*,\s*([0-9\.]+)\s*\)')

    def __init__(self, r, g, b, a=1.0):
        self._r = float(r)
        self._g = float(g)
        self._b = float(b)
        self._a = float(a)

    def copy(self):
        return Color(self._r, self._g, self._b, self._a)

    @classmethod
    def construct(cls, context, obj):
        if isinstance(obj, cls):
            return obj.copy()
        if isinstance(obj, text_type):
            return cls.parse(obj)
        if isinstance(obj, (float, int)):
            return cls(obj, obj, obj)

        try:
            r, g, b, a = obj
        except:
            pass
        else:
            return cls(r, g, b, a)

        try:
            r, g, b = obj
        except:
            pass
        else:
            return cls(r, g, b)

        raise ValueError("unable to convert {} to a color".format(context.to_expr(obj)))

    @classmethod
    def parse(cls, txt):
        txt = txt.strip().lower()
        if txt == 'inherit':
            return 'inherit'
        if txt in HTML_COLORS:
            return HTML_COLORS[txt]
        elif txt.startswith('#'):
            return cls.parse_hex(txt)
        elif txt.startswith('rgba'):
            return cls.parse_rgba(txt)
        elif txt.startswith('rgb'):
            return cls.parse_rgb(txt)
        elif txt.startswith('hsla'):
            return cls.parse_hsla(txt)
        elif txt.startswith('hsl'):
            return cls.parse_hsl(txt)
        else:
            raise ValueError("unable to parse '{}' as a color".format(txt))

    @classmethod
    def parse_hex(cls, hex):
        match = cls._re_hex_short.match(hex)
        if match is not None:
            r, g, b = [int(c + c, 16) for c in match.groups()]
            return cls(r, g, b)
        else:
            match = cls._re_hex.match(hex)
            if match is None:
                raise ValueError('not valid hex color')
            r, g, b = [int(c, 16) for c in match.groups()]
            return cls(r, g, b)

    @classmethod
    def parse_rgb(cls, rgb):
        match = cls._re_rgb.match(rgb)
        if match is None:
            raise ValueError('not a valid rgb color')
        r, g, b = [int(c) for c in match.groups()]
        return cls(r, g, b)

    @classmethod
    def parse_rgba(cls, rgba):
        match = cls._re_rgba.match(rgba)
        if match is None:
            raise ValueError('not a valid rgb color')
        r, g, b = [int(c) for c in match.groups()[:3]]
        a = float(match.groups()[3])
        return cls(r, g, b, a)

    @classmethod
    def parse_hsl(cls, hsl):
        match = cls._re_hsl.match(hsl)
        if match is None:
            raise ValueError('not a valid hsl color')
        h, s, l = [float(c) for c in match.groups()]
        h = h / 360.0
        l = l / 100.0
        s = s / 100.0
        r, g, b = colorsys.hls_to_rgb(h, l, s)
        return cls(r * 255.0, g * 255.0, b * 255.0)

    @classmethod
    def parse_hsla(cls, hsla):
        match = cls._re_hsla.match(hsla)
        if match is None:
            raise ValueError('not a valid hsla color')
        h, s, l, a = [float(c) for c in match.groups()]
        h = h / 360.0
        l = l / 100.0
        s = s / 100.0
        r, g, b = colorsys.hls_to_rgb(h, l, s)
        return cls(r * 255.0, g * 255.0, b * 255.0, a * 255.0)

    def __moyaconsole__(self, console):
        from moya import pilot
        link = "http://www.color-hex.com/color/{}".format(self.hex.lower()[1:])
        mr = self.__moyarepr__(pilot.context)
        console.text(mr, bold=True)
        console.text(link, fg="cyan", underline=True)
        # keys = OrderedDict((k, getattr(self, k)) for k in self.__moya_exposed_attributes__)
        # console.obj(pilot.context, keys)

    def __str__(self):
        return self.html

    def __repr__(self):
        return "Color({!r}, {!r}, {!r}, {!r})".format(self._r, self._g, self._b, self._a)

    def __moyarepr__(self, context):
        return "color:[{:g}, {:g}, {:g}, {:g}]".format(self._r, self._g, self._b, self._a)

    def __hash__(self):
        return hash((self._r, self._g, self._b, self._a))

    def as_pillow_tuple(self):
        if self.a == 1.0:
            return (int(self.r), int(self.g), int(self.b))
        else:
            return (int(self.r), int(self.g), int(self.b), int(self.a * 255.0))

    @property
    def html(self):
        if self.a == 1.0:
            return self.hex
        else:
            return self.rgba

    @property
    def hex(self):
        if any(c % 17 for c in self._rgb):
            return "#{:02X}{:02X}{:02X}".format(*self._rgb)
        else:
            return "#{:0X}{:0X}{:0X}".format(*(c // 17 for c in self._rgb))

    @property
    def rgb(self):
        return "rgb({},{},{})".format(*self._rgb)

    @property
    def rgba(self):
        return "rgba({},{},{},{:g})".format(*self._rgba)

    @property
    def hsl(self):
        h, l, s = colorsys.rgb_to_hls(self.r / 255.0, self.g / 255.0, self.b / 255.0)
        return "hsl({:g},{:g}%,{:g}%)".format(h * 360.0, s * 100.0, l * 100.0)

    @property
    def hsla(self):
        h, l, s = colorsys.rgb_to_hls(self.r / 255.0, self.g / 255.0, self.b / 255.0)
        return "hsla({:g},{:g}%,{:g}%,{:g})".format(h * 360.0, s * 100.0, l * 100.0, self.a)

    @property
    def _rgb(self):
        return [int(self.r), int(self.g), int(self.b)]

    @property
    def _rgba(self):
        return [int(self.r), int(self.g), int(self.b), self.a]

    @property
    def opaque(self):
        return Color(self._r, self._g, self._b, 1.0)

    @property
    def saturated(self):
        return Color(self.r, self.g, self.b, self.a)

    @property
    def r(self):
        return min(255.0, max(0.0, self._r))

    @property
    def g(self):
        return min(255.0, max(0.0, self._g))

    @property
    def b(self):
        return min(255.0, max(0.0, self._b))

    @property
    def a(self):
        return min(1.0, max(0.0, self._a))

    @property
    def name(self):
        return HTML_COLORS_REVERSE.get(self, None)

    def __eq__(self, other):
        if not isinstance(other, Color):
            raise ValueError('can only compare a color to another color')
        return self._r == other._r and self._g == other._g and self._b == other._b and self._a == other._a

    def __add__(self, other):
        if not isinstance(other, Color):
            raise ValueError('can only add another color to a color')
        r, g, b, a = self._rgba
        _r, _b, _g, _a = other._rgba
        return Color(r + _r, g + _g, b + _b, a)

    def __sub__(self, other):
        if not isinstance(other, Color):
            raise ValueError('can only subtract another color from a color')
        r, g, b, a = self._rgba
        _r, _b, _g, _a = other._rgba
        return Color(r - _r, g - _g, b - _b, a)

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            r, g, b, a = self._rgba
            return Color(r * other, g * other, b * other, a)
        if not isinstance(other, Color):
            raise ValueError('can only multiple a color by a number or other color')
        r, g, b, a = self._rgba
        _r, _g, _b, _a = other._rgba
        return Color(r * _r, g * _g, b * _b, a)

    def __div__(self, other):
        raise NotImplementedError("can't divide a color")

HTML_COLORS = {k: Color.parse(v) for k, v in HTML_COLORS.items()}
HTML_COLORS_REVERSE = {v: k for k, v in HTML_COLORS.items()}


if __name__ == "__main__":
    print(unicode(Color(20, 255, 0)))
    print(unicode(Color(20, 255, 0, 0.5)))

    print(unicode(Color.parse_hex('#00FF34').html))

    print(unicode(Color.parse_rgb('rgb(100, 23,50)').html))

    print(unicode(Color.parse_rgba('rgba(100, 23,50, 0.6)').html))

    print(Color.parse('tomato'))
    print(Color.parse('white'))

    print(Color.parse('hsl(200, 50%, 10%)'))
    print(Color.parse('hsla(200, 50%, 10%, 0.5)'))

    print(Color.parse('hsl(120, 100%, 50%)'))



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from . import pilot
from .containers import LRUCache
from .syntax import highlight
from .compat import implements_to_string, implements_bool, text_type, iteritems
from .console import ConsoleHighlighter
from .context.tools import to_expression

import re
from collections import namedtuple, defaultdict
from operator import attrgetter
from .compat import quote

from pyparsing import Literal, QuotedString, Word, OneOrMore, printables
import pyparsing
pyparsing.ParserElement.enablePackrat()


class URLHighlight(ConsoleHighlighter):
    styles = {
        None: "bold blue",
        "asterix": "bold yellow",
        "param": "bold magenta"
    }

    highlights = [
        r'(?P<param>\{.*?\})',
        r'(?P<asterix>\*)',
    ]


class TargetHighlight(ConsoleHighlighter):
    styles = {
        "lib": "bold blue",
        "name": "bold green",
        "hash": "white"
    }

    highlights = [
        r'(?P<lib>.*?)(?P<hash>#)(?P<name>.*?)[,\s$]',
    ]


class RouteError(ValueError):
    hide_py_traceback = True


class DuplicatedParameter(RouteError):
    """a url parameter was repeated"""


class MissingURLParameter(RouteError):
    """A parameter in a route was not supplied"""


class NamedURLNotFound(RouteError):
    """No such named route exists"""


def _router_grammer():
    """Make the url route grammer"""
    wildcard = Literal('*')
    path_separator = Literal('/').setParseAction(lambda t: ('sep', t))
    extract = QuotedString(quoteChar='{', endQuoteChar='}').setParseAction(lambda t: ('extract', t))
    static = Word(''.join(c for c in printables if c not in '?{}/')).setParseAction(lambda t: ('static', t))
    route_parse = OneOrMore(extract | static | path_separator | wildcard)
    return route_parse

RouteMatch = namedtuple("RouteMatch", ["data", "target", "name"])


@implements_to_string
@implements_bool
class URLProxy(object):

    def __init__(self, mapper, path=None):
        self.mapper = mapper
        if path is None:
            path = []
        self.path = path
        self.bound = None

    @classmethod
    def url_join(self, url_parts, from_root=False, _consecutive_slashes_sub=re.compile(r'/+').sub):
        url = _consecutive_slashes_sub('/', ''.join(url_parts))
        if from_root and not url.startswith('/'):
            return '/' + url
        return url

    def bind(self, data):
        self.bound = data

    def __getitem__(self, name):
        obj = self.evaluate()
        if isinstance(obj, text_type):
            return obj[name]
        if name not in obj.named_routes:
            raise KeyError(name)
        proxy = URLProxy(self.mapper, self.path + [name])
        #obj = proxy.evaluate()
        #if isinstance(obj, text_type):
        #    return obj
        return proxy

    def __moyacall__(self, params):
        self.bound = params
        return self

    def __str__(self):
        obj = self.evaluate()
        if isinstance(obj, text_type):
            return obj
        else:
            return self.base_url

    def __repr__(self):
        return "<urlproxy>"

    def __moyarepr__(self, context):
        obj = self.evaluate()
        if isinstance(obj, text_type):
            return to_expression(context, obj)
        base_url = self.base_url
        if base_url is None:
            return "<mapper>"
        else:
            return "<mapper '{}'>".format(base_url)

    def keys(self):
        obj = self.evaluate()
        if isinstance(obj, text_type):
            return []
        return sorted(obj.named_routes.keys())

    def values(self):
        return [self[k] for k in self.keys()]

    def items(self):
        return zip(self.keys(), self.values())

    def __len__(self):
        return len(self.keys())

    def __contains__(self, k):
        obj = self.evaluate()
        if isinstance(obj, text_type):
            return k in obj
        return k in obj.named_routes

    # This class is a bit schitzo; it can behave like a list or a string
    def __bool__(self):
        obj = self.evaluate()
        if isinstance(obj, text_type):
            return bool(obj)
        return bool(obj.named_routes)

    @property
    def base_route(self):
        mapper = self.mapper
        url_route = []
        for name in self.path:
            route = mapper.named_routes[name][0]
            url_route.append(route._route)
            if isinstance(route.target, URLMapper):
                mapper = route.target
            else:
                break
        return self.url_join(url_route, from_root=True)

    @property
    def base_url(self):
        if self.bound is None:
            data = pilot.context
        else:
            data = self.bound
        mapper = self.mapper
        url_parts = []
        for name in self.path:
            route = mapper.named_routes[name][0]
            if isinstance(route.target, URLMapper):
                url_parts.append(route.generate(data)[1])
                mapper = route.target
            else:
                break
        return self.url_join(url_parts, from_root=True)

    def evaluate(self):
        if self.bound is None:
            data = pilot.context
        else:
            data = self.bound
        mapper = self.mapper
        url_parts = []
        for name in self.path:
            route = mapper.named_routes[name][0]
            if isinstance(route.target, URLMapper):
                url_parts.append(route.generate(data)[1])
                mapper = route.target
            else:
                url_parts.append(route.generate(data)[1])
                url = self.url_join(url_parts, from_root=True)
                return url
        return mapper

    def __moyaconsole__(self, console):
        _mapper = self.evaluate()
        if isinstance(_mapper, text_type):
            console.obj(pilot.context, _mapper)
            return

        table = []
        console("URL mapper on ", fg="blue", bold=True)(text_type(self.base_route), bold=True, fg="cyan").nl()
        for mapper in _mapper.routes:
            name = mapper.name
            if isinstance(mapper.target, (list, tuple)):
                target_type = "final"
                str_target = ", ".join(mapper.target)
                mount = mapper.route
            else:
                mount = mapper.route
                target_type = "mountpoint"
                str_target = ''
            table.append([name or '', target_type, self.url_join([self.base_url, mount]), str_target])
        console.table(table,
                      header_row=["name", "type", "route", "target"],
                      cell_processors={2: URLHighlight.highlight, 3: TargetHighlight.highlight})

    def __moyarenderable__(self, context):
        request = context['.request']
        renderable = []

        def recurse_urlmapper(mapper, url_parts=None, app=None):
            if url_parts is None:
                url_parts = ['/']
            for mapper in mapper.routes:
                if not mapper.match_method(request.method, None):
                    continue
                methods = ', '.join(mapper.methods or ['*'])

                if isinstance(mapper.target, (list, tuple)):
                    mount = mapper.route
                    url = self.url_join(url_parts + [mount])
                    url = highlight("route", url, line_numbers=False)

                    targets = [(target, highlight("target", target, line_numbers=False))
                               for target in mapper.target]

                    renderable.append([url, methods, app, targets])
                else:
                    url = self.url_join(url_parts + [mapper.route])
                    _, params = mapper.partial_parse(url)
                    app = params.get('app', app) if params else None
                    recurse_urlmapper(mapper.target, url_parts + [mapper.route], app=app)

        recurse_urlmapper(self.mapper)
        return renderable


class _NamedURLs(object):
    """Retrieves named urls a mapper and its descendants"""

    def __init__(self, mapper, cache_size=10000):
        self.mapper = mapper
        self.named_routes = self._get_named_routes()
        self.cache = LRUCache(cache_size)

    def __repr__(self):
        return "<namedurls '{}'>".format(self.mapper.name)

    def _get_named_routes(self):
        named_routes = defaultdict(list)

        def recurse_mapper(mapper, path):
            for route in mapper.routes:
                target = route.target
                if isinstance(target, URLMapper):
                    recurse_mapper(target, path + (route,))
                else:
                    if route.name is not None:
                        named_routes[route.name].append(path + (route,))
        recurse_mapper(self.mapper, ())

        return named_routes

    def get_url(self, name, params=None, route=None):
        if params is None:
            params = {}
        cache_key = (id(route), name, tuple(sorted(params.items())))
        cached_url = self.cache.get(cache_key, None)
        if cached_url is not None:
            return cached_url

        matches = []
        for i, path in enumerate(self.named_routes[name]):
            if route is not None:
                path = (route,) + path
            try:
                remaining_params, url = self._reverse_url_path(path, params)
            except MissingURLParameter:
                # This path requires a url parameter that we don't have
                continue
            else:
                matches.append((remaining_params, i, url))

        if matches:
            matches.sort()
            url = matches[0][-1]
            self.cache[cache_key] = url
            return url

        if params:
            params_display = ",".join("{}='{}'".format(k, v) for k, v in iteritems(params))
            raise NamedURLNotFound("No matching URL with name '{}' and parameters {}".format(name, params_display))
        else:
            raise NamedURLNotFound("No matching URL with name '{}' and no parameters".format(name))

    def _reverse_url_path(self, path, params, _consecutive_slashes_sub=re.compile(r'/+').sub):
        url_components = []
        path_matched = set()
        for route in path:
            matched, url = route.generate(params)
            path_matched.update(matched)
            url_components.append(url)
        url = _consecutive_slashes_sub('/', ''.join(url_components))
        if not url.startswith('/'):
            url = '/' + url
        return len(params) - len(path_matched), url


class URLMapper(object):

    def __init__(self, name=''):
        self.name = name
        self._routes = []
        self._route_cache = LRUCache(1000)
        self._named_routes = defaultdict(list)
        self._named_urls = None
        self._proxy = None
        self._insert_order = 0
        self._finalized = False

    def __repr__(self):
        if self.name:
            return "<urlmapper '{}'>".format(self.name)
        else:
            return "<urlmapper>"

    def __moyacontext__(self, context):
        if self._proxy is None:
            self._proxy = URLProxy(self)
        return self._proxy

    @property
    def routes(self):
        self.finalize()
        return self._routes

    @property
    def named_routes(self):
        self.finalize()
        return self._named_routes

    def _get_order(self, priority):
        """Get a tuple for use as a route order attribute"""
        order = (-priority, self._insert_order)
        self._insert_order += 1
        return order

    def finalize(self, _get_order=attrgetter('order')):
        """Finalize the url mapper (does sorting)"""
        if not self._finalized:
            self._finalized = True
            self._routes.sort(key=_get_order)
            for k, v in iteritems(self._named_routes):
                v.sort(key=_get_order)

    @property
    def named_urls(self):
        if self._named_urls is None:
            self._named_urls = _NamedURLs(self)
        return self._named_urls

    def map(self, url, target, methods=None, handlers=None, defaults=None, name=None, priority=0):
        """Map a url on to a target object"""
        route = Route(self,
                      url,
                      partial=False,
                      defaults=defaults or {},
                      methods=methods,
                      handlers=handlers,
                      target=target,
                      name=name,
                      order=self._get_order(priority))
        route.re_route
        self._routes.append(route)
        if name is not None:
            self._named_routes[name].append(route)
        self._finalized = False
        return route

    def mount(self, url, mapper, defaults=None, name=None, priority=0):
        """Mount a sub-mapper on a url"""
        if url:
            if not url.startswith('/'):
                url = '/' + url
            if not url.endswith('/'):
                url += '/'
        route = Route(self,
                      url,
                      partial=True,
                      defaults=defaults or {},
                      target=mapper,
                      methods=None,
                      name=name,
                      order=self._get_order(priority))
        if name is not None:
            self._named_routes[name].append(route)
        self._routes.append(route)
        self._finalized = False
        return route

    def get_routes(self, name, default=Ellipsis):
        routes = self.named_routes.get(name, default)
        if routes is Ellipsis:
            raise RouteError("Named route '{}' not found".format(name))
        return routes

    def get_route(self, url, method="GET", handler=None):
        """Process a URL for any corresponding routes"""
        for route_match in self.iter_routes(url, method, handler):
            return route_match
        return None

    def iter_routes(self, url, method="GET", handler=None):
        """Yield any routes that match the url"""

        route_key = (url, method, handler)

        if route_key in self._route_cache:
            route_matches = self._route_cache.lookup(route_key)

        else:
            self.finalize()
            route_matches = []
            add_route_match = route_matches.append
            for route in self._routes:
                if route.match_method(method, handler):
                    if route.partial:
                        remaining_url, url_data = route.partial_parse(url)
                        if remaining_url is None:
                            continue
                        if not remaining_url.startswith('/'):
                            remaining_url = '/' + remaining_url
                        sub_url_router = route.target
                        for route_match in sub_url_router.iter_routes(remaining_url,
                                                                      method=method,
                                                                      handler=handler):
                            sub_url_data = url_data.copy()
                            sub_url_data.update(route_match.data)
                            add_route_match(RouteMatch(sub_url_data, route_match.target, route_match.name))
                    else:
                        route_data = route.parse(url)
                        if route_data is not None:
                            add_route_match(RouteMatch(route_data, route.target, route.name))

            self._route_cache[route_key] = route_matches
        return iter(route_matches)

    def has_route(self, url, method="GET", handler=None):
        for route in self.iter_routes(url, method, handler):
            return True
        return False

    def get_url(self, name, params=None, base_route=None):
        return self.named_urls.get_url(name, params, route=base_route)

    def render(self, level=0):
        for route in self.routes:
            print('    ' * level + repr(route))
            if isinstance(route.target, URLMapper):
                route.target.render(level + 1)


@implements_to_string
class Route(object):
    """Parses URLs and extracts parameters from them"""

    _route_parse = _router_grammer()

    _patterns = {'': (r"[^/]+?", lambda s: s),
                 '*': (r".*?", lambda s: s),
                 'alpha': (r"[a-zA-Z0-9]+?", lambda s: s),
                 'slug': (r"[\w_-]+?", lambda s: s),

                 "integer": (r"\-?\d+?", int),
                 "float": (r"\-?\d+\.\d+", float),

                 "posinteger": (r"\d+?", int),
                 "posfloat": (r"\d+\.\d+", float)
                 }

    def __init__(self,
                 mapper,
                 route,
                 partial=False,
                 defaults={},
                 methods=None,
                 handlers=None,
                 target=None,
                 name=None,
                 order=(0, 0)):
        self.mapper = mapper
        self._route = route
        self.route = route
        self.partial = partial
        self.defaults = defaults
        self.methods = methods
        self.handlers = handlers or []
        self.target = target
        self.name = name
        self.order = order

        self.component_names = []
        self.component_callables = {}
        self.handlers = [int(handler) for handler in self.handlers]
        self._tokens = None
        self._re_route = None

    @property
    def tokens(self):
        if self._tokens is None:
            self._tokens = ([(tt, t[0])
                            for (tt, t) in self._route_parse.parseString(self._route, True)]
                            if self._route else [])
        return self._tokens

    @classmethod
    def _split_pattern_name(cls, token):
        if ':' in token:
            return token.split(':', 1)
        return '', token

    @property
    def re_route(self):
        if not self._re_route:
            segments = ['^']
            append_segment = segments.append
            escape = re.escape
            static = True
            default_pattern = self._patterns.get('')
            for token_type, token in self.tokens:
                if token_type == 'extract':
                    static = False
                    if token.startswith('*'):
                        pattern_name, name = self._split_pattern_name(token[1:])
                        pattern, component_callable = self._patterns.get(pattern_name or '*', default_pattern)
                        if name:
                            if name in self.component_names:
                                raise DuplicatedParameter("URL parameter '{}' was duplicated".format(name))
                            append_segment('(?P<%s>%s)' % (name, pattern))
                        else:
                            append_segment('(.*)')
                        self.component_names.append(name)
                    else:
                        pattern_name, name = self._split_pattern_name(token)
                        pattern, component_callable = self._patterns.get(pattern_name, default_pattern)
                        append_segment('(?P<%s>%s)' % (name, pattern))
                        if name in self.component_names:
                            raise DuplicatedParameter("URL parameter '{}' was duplicated".format(name))
                        self.component_names.append(name)
                    self.component_callables[name] = component_callable
                else:
                    if token == '*':
                        append_segment('.*?')
                    else:
                        append_segment(escape(token))

            if not self.partial:
                append_segment('$')

            self.static = static
            self.re_route_text = ''.join(segments)
            self._re_route = None
            self._re_route = re.compile(self.re_route_text)

        return self._re_route

    def __str__(self):
        if isinstance(self.target, tuple):
            target = ", ".join(self.target)
        else:
            target = str(self.target)
        if self.handlers:
            m = ",".join(text_type(h) for h in self.handlers)
        else:
            m = ','.join(self.methods or ['*'])

        return "<route '%s' %s %s: %s>" % (self.name, m, self.route, target)

    def __repr__(self):
        if isinstance(self.target, tuple):
            target = ", ".join(self.target)
        else:
            target = str(self.target)
        m = ','.join(self.methods or ['*'])
        return "<route '%s' %s %s:%s>" % (self.name, m, self.route, target)

    def update_defaults(self, defaults):
        """Update URL route defaults"""
        self.defaults.update(defaults)

    def match_method(self, method, handler):
        if self.handlers:
            if handler is None:
                return False
            return handler in self.handlers
        if handler:
            if self.partial:
                return True
            return False
        return (self.methods is None or
                method == '*' or
                method in self.methods)

    def parse(self, url):
        """Parse a URL and return a dictionary containing parameters"""
        match = self.re_route.match(url)
        null_callable = lambda v: v
        get_component_callable = self.component_callables.get
        if match:
            m = self.defaults.copy()
            m.update(match.groupdict())
            try:
                m = dict((k, get_component_callable(k, null_callable)(v))
                         for k, v in m.items())
            except ValueError:
                return None
            return m
        return None

    def partial_parse(self, url):
        """Parses a URL as far as possible and returns a tuple of the remaining portion of the URL
        and the parsed parameters.

        """
        params = self.parse(url)
        if params is None:
            return None, None
        params = params or {}
        partial_url = self.generate(params)[1]
        remaining_url = url[len(partial_url):]
        return remaining_url, params

    def generate(self, component_map, remove=False):
        """Recreate a URL from the given parameters"""
        self.re_route
        if self.static:
            return set(), self.route

        matched = set()
        segments = []
        append = segments.append
        url_params = self.defaults.copy()

        for name in self.component_names:
            if name in component_map:
                matched.add(name)
                url_params[name] = component_map[name]

        for token_type, token in self.tokens:
            if token_type == "extract":
                try:
                    name = token.lstrip('*')
                    if ':' in name:
                        _, name = name.split(':', 1)
                    append(url_params[name])
                except KeyError:
                    raise MissingURLParameter("Can't generate URL without a value for '{}'".format(name))
            else:
                append(token)
        return matched, ''.join(quote(text_type(s)) for s in segments)

    # def reverse_url(self, params):
    #     components = []
    #     route = self
    #     while route:
    #         components.append(route.generate(params))
    #         if route.mapper is None:
    #             break
    #         route = route.mapper.route
    #     url = ''.join(reversed(components)).replace('//', '/')
    #     if not url.startswith('/'):
    #         url = '/' + url
    #     return url

    # def reverse_url_path(self, path, params):
    #     components = [route.generate(params) for route in path]
    #     url = ''.join(components).replace('//', '/')
    #     if not url.startswith('/'):
    #         url = '/' + url
    #     return url


if __name__ == "__main__":
    mapper = URLMapper('server')

    uploads_mapper = URLMapper('uploads')
    media_mapper = URLMapper('media')

    static_mapper = URLMapper('static')
    static_mapper.map('/{*path}/', '#server', name="serve")

    uploads_mapper.mount('/uploads/', static_mapper, name="uploads")
    media_mapper.mount('/static/', static_mapper, name="static")

    uploads_mapper.render()
    media_mapper.render()

    print(uploads_mapper.get_url('serve', {'path': "css/content.css"}))
    #import rpdb2; rpdb2.start_embedded_debugger('password')
    print(media_mapper.get_url('serve', {'path': "css/content.css"}))



from __future__ import unicode_literals
from __future__ import print_function

from time import time
from datetime import datetime
from contextlib import contextmanager
from random import choice
from functools import reduce
from calendar import timegm

from os.path import abspath, dirname, join, isfile, basename
import re
import os
import sys
import hashlib

from .compat import zip_longest, iteritems, xrange, text_type, PY3, implements_to_string, number_types


_re_xml_namespace = re.compile(r'^\{(.*?)\}(.*)$', re.UNICODE)


def extract_namespace(tag_name, default='http://moyaproject.com', _namespace_match=_re_xml_namespace.match):
    """Extracts namespace and tag name in Clark's notation"""
    match = _namespace_match(tag_name)
    if match is None:
        return default, tag_name
    return match.group(1) or default, match.group(2)


def asint(value, default=None):
    """Convert to an integer where possible, or return a default"""
    try:
        return int(value)
    except:
        return default


def match_exception(exception, catch_type):
    """Match an exception with a catch type wildcard"""
    for c, e in zip_longest(catch_type.split('.'),
                            exception.split('.'),
                            fillvalue=None):
        if c == '*':
            return True
        if c != e:
            return False
    return True


def md5_hexdigest(text):
    m = hashlib.md5()
    m.update(text.encode('utf-8'))
    if PY3:
        return m.hexdigest()
    else:
        return m.hexdigest().decode('utf-8')


def check_missing(map):
    for k, v in iteritems(map):
        if getattr(v, 'moya_missing', False):
            raise ValueError("value '{}' must not be missing (it is {!r})".format(k, v))


@contextmanager
def timer(msg="elapsed", ms=False, write_file=None, output=sys.stdout.write):
    """Context manager to time a block of code"""
    now = datetime.now()
    start = time()
    yield
    taken = (time() - start)
    if ms:
        output("%s: %.2fms\n" % (msg, taken * 1000))
    else:
        output("%s: %.2fs\n" % (msg, taken))
    if write_file is not None:
        import socket
        hostname = socket.gethostname()
        with open(write_file, 'a+') as f:
            f.write('%s,%.2f,"%s","%s"\n' % (now.ctime(), taken, msg, hostname))


class TimeDeltaParser(object):
    """Convert a text description of a time span in to milliseconds"""
    _re_td = re.compile(r'(\d+?)(ms|s|m|h|d)?$')
    _to_ms = dict(ms=1,                   # milliseconds
                  s=1000,                 # seconds
                  m=1000 * 60,            # minutes
                  h=1000 * 60 * 60,       # hours
                  d=1000 * 60 * 60 * 24)  # days

    @classmethod
    def parse(cls, s):
        """Convert a timedelta string to an integer of milliseconds,
        or return None if the string is not in the appropriate format.

        """
        match = cls._re_td.match(s)
        if match is None:
            raise ValueError("'{}' is not a valid timespan".format(s))
        t, unit = match.groups()
        t = int(t) * cls._to_ms.get(unit or 'ms')
        return t

parse_timedelta = TimeDeltaParser.parse


def get_moya_dir(path=None):
    """Searches current directory and ancestors for a directory containing
    a file called 'moya' -- or None if no path is found.

    """
    if path is None:
        path = os.getcwd()
    path = abspath(path)
    while not isfile(join(path, 'moya')):
        if basename(path) in ('', '/'):
            raise ValueError("Moya project directory not found")
        path = dirname(path)
    return path


def is_moya_dir(path=None):
    """Check if a path is a moya project"""
    if path is None:
        path = os.curdir
    path = abspath(path.replace('\\', '/'))
    return isfile(join(path, 'moya'))


def file_chunker(f, size=2 * 16 * 1024):
    """An iterator that reads a file in chunks"""
    read = f.read
    chunk = read(size)
    while chunk:
        yield chunk
        chunk = read(size)
    f.close()


def make_id():
    """Make a unique id"""
    _ID_CHARS = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    t = "%0X" % time()
    return "%s%s" % (t, ''.join(choice(_ID_CHARS) for _ in xrange(6)))


def datetime_to_epoch(d):
    if isinstance(d, number_types):
        return int(d)
    return timegm(d.utctimetuple())


def split_commas(s):
    """Split a string on commas and return a list of tokens"""
    return [token.strip() for token in s.split(',') if token]


def summarize_text(text, max_length=100, marker='[...]'):
    """Truncate text and add a marker, if the length is above `max_langth`"""
    if not text:
        return ''
    if len(text) < max_length:
        return text
    return text[:max_length] + marker


def get_return(value):
    if value is None:
        return {}
    if hasattr(value, 'get_return_value'):
        return value.get_return_value()
    return value


def as_dict(value):
    if isinstance(value, dict):
        return value
    return dict(iteritems(value))


def quote(text):
    """Return a string surrounded by quotes"""
    # This may be trivial, but it can make code more self-documenting
    return '"{}"'.format(text)


def squote(text):
    """Return a string surrounded by single quotes"""
    return "'{}'".format(text)


def textual_list(items, join_word='or', empty="(nothing)"):
    """Lists items in humanized list form.

    e.g.
    >>> textual_list(['foo', 'bar', 'baz'])
    'foo', 'bar' or 'baz'

    """
    if not items:
        return empty
    items = list(items)
    last_item = items.pop()
    items_list = ", ".join("'{}'".format(item) for item in items)
    if items:
        return "{} {} {}".format(items_list, join_word, "'{}'".format(last_item))
    else:
        return "'{}'".format(last_item)


def moya_update(map, values):
    (getattr(map, '__moyaupdate__', None) or getattr(map, 'update'))(values)


def url_join(*urls):
    """Combine url segments"""
    return reduce(lambda p1, p2: '{}/{}'.format(p1.rstrip('/'), p2.lstrip('/')), urls)


def remove_padding(text):
    """remove any blank (all whitespace) lines at the beginning and end of text"""
    if text.isspace() or not text:
        return ''
    lines = text.splitlines()
    padding_start = 0
    padding_end = 0
    iter_lines = iter(lines)
    while (next(iter_lines) or ' ').isspace():
        padding_start += 1
    iter_lines = iter(reversed(lines))
    while (next(iter_lines) or ' ').isspace():
        padding_end -= 1
    return '\n'.join(lines[padding_start or None:padding_end or None])


def unique(v):
    """Generates a list from a sequence where each value appears only once, and the order is preserved."""
    try:
        sequence = list(v)
    except:
        return []
    seen = set()
    seen_add = seen.add
    return [item for item in sequence if not (item in seen or seen_add(item))]


def format_element_type(element_type):
    if isinstance(element_type, text_type):
        return element_type
    ns, et = element_type
    return "{" + ns + "}" + et


# def get_ids(seq):
#     """Get ID attributes from a sequence"""
#     return [item.id for item in seq if hasattr(seq, 'id')]


class MultiReplace(object):
    """Replace multiple tokens at once"""
    # Note, this typically slower than chained calles to .replace
    def __init__(self, replace_map):
        self.get_replace = replace_map.__getitem__
        self._re = re.compile("|".join(re.escape(k) for k in replace_map.keys()))
        self.sub = self._re.sub

    def __call__(self, text):
        get_replace = self.get_replace
        return self.sub(lambda match: get_replace(match.group(0)), text)


class DummyLock(object):
    """Replacement for real lock that does nothing"""
    def __enter__(self):
        pass

    def __exit__(self, *args, **kwargs):
        pass


def make_cache_key(key_data):
    if not isinstance(key_data, (list, tuple)):
        key_data = [key_data]
    key = []
    append = key.append
    for component in key_data:
        if isinstance(component, text_type):
            append(component)
        elif isinstance(component, list):
            append('-'.join(make_cache_key(k) for k in component))
        elif isinstance(component, set):
            append('-'.join(make_cache_key(k) for k in sorted(component)))
        elif isinstance(component, dict):
            append('-'.join("{}_{}".format(k, make_cache_key([v])) for k, v in sorted(component.items())))
        else:
            append(text_type(component))
    return '.'.join(key)


# http://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Levenshtein_distance#Python
def levenshtein(seq1, seq2):
    """Levenshtein word distance"""
    oneago = None
    thisrow = list(range(1, len(seq2) + 1)) + [0]
    for x in xrange(len(seq1)):
        twoago, oneago, thisrow = oneago, thisrow, [0] * len(seq2) + [x + 1]
        for y in xrange(len(seq2)):
            delcost = oneago[y] + 1
            addcost = thisrow[y - 1] + 1
            subcost = oneago[y - 1] + (seq1[x] != seq2[y])
            thisrow[y] = min(delcost, addcost, subcost)
    return thisrow[len(seq2) - 1]


def nearest_word(word, words, min_distance=3):
    """Find the nearest word from a list, within a given Levenshtein distance"""
    compare = []
    for compare_word in words:
        d = levenshtein(word, compare_word)
        if d < min_distance:
            compare.append((d, compare_word))
    if not compare:
        return None
    return sorted(compare)[0][1]


def show_tb(f):
    """Decorator that print tracebacks from functions that may otherwise end up hiding them"""
    def deco(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            import traceback
            traceback.print_exc(e)
            raise
    return deco


def normalize_url_path(p):
    """Makes sure a URL path starts and ends with a forward slash"""
    if not p.startswith('/'):
        p = '/' + p
    if not p.endswith('/'):
        p = p + '/'
    return p


@implements_to_string
class lazystr(object):
    """convert to a string lazily, for use in logging"""
    def __init__(self, _callable, *args, **kwargs):
        self._callable = _callable
        self._args = args
        self._kwargs = kwargs
        self._str = None

    def __str__(self):
        if self._str is None:
            self._str = text_type(self._callable(*self._args, **self._kwargs))
        return self._str

    def __len__(self):
        return len(text_type(self))

    def __getattr__(self, k):
        return getattr(text_type(self), k)




from __future__ import unicode_literals
from __future__ import print_function

import re

from .tools import timer, remove_padding
from .compat import iteritems, with_metaclass
from .filter import MoyaFilterBase
from .render import HTML
from .compat import text_type

from textwrap import dedent


def _escape_html(text):
    """Escape text for inclusion in html"""
    return text.replace('&', '&amp;').replace('<', '&lt;').replace('>', '&gt;').replace(' ', '&nbsp;')


def tabs_to_spaces(line, tab_size=4):
    """Converts tabs to a fixed numbers of spaces at the beginning of a string"""
    spaces = 0
    for c in line:
        if c not in ' \t':
            break
        if c == '\t':
            spaces += tab_size - (spaces % tab_size)
        else:
            spaces += 1
    return ' ' * spaces + line.lstrip()


def highlight(format,
              code,
              start_line=1,
              end_line=None,
              line_numbers=True,
              highlight_lines=None,
              highlight_range=None,
              highlight_range_style="error"):
    if isinstance(code, bytes):
        code = code.decode('utf-8', 'replace')

    HL = HighlighterMeta.highlighters.get(format, Highlighter)
    h = HL()
    html = h.highlight(code,
                       start_line,
                       end_line,
                       line_numbers=line_numbers,
                       highlight_lines=highlight_lines,
                       highlight_range=highlight_range,
                       highlight_range_style=highlight_range_style)
    return html


class HighlighterMeta(type):
    highlighters = {}

    def __new__(cls, name, base, attrs):
        new_class = type.__new__(cls, name, base, attrs)
        format = getattr(new_class, 'format', None)
        if format:
            cls.highlighters[format] = new_class
        flags = re.UNICODE | re.MULTILINE | re.DOTALL
        new_class._compiled_styles = [re.compile(s, flags) for s in new_class.styles]
        return new_class


class HighlighterType(object):
    """Somewhat naive syntax highlighter"""
    line_anchors = True

    styles = []
    _compiled_styles = None

    _re_linebreaks = re.compile(r'$', flags=re.UNICODE | re.MULTILINE | re.DOTALL)

    _highlight_range_padding = 50

    def highlight(self,
                  code,
                  start_line=None,
                  end_line=None,
                  line_numbers=True,
                  highlight_lines=None,
                  highlight_range=None,
                  highlight_range_style="error"):
        if start_line is None:
            start_line = 1

        offset_line = 0

        start_line = max(1, start_line)
        lines = code.splitlines()
        if end_line is None:
            end_line = len(lines)
        lines = ([''] * offset_line) + lines[offset_line:end_line + self._highlight_range_padding + 1]
        code = '\n'.join(tabs_to_spaces(l.rstrip()) for l in lines)

        points = []
        line_starts = [-1]
        add_point = points.append

        for style_regex in self._compiled_styles:
            for match in style_regex.finditer(code):
                for k, v in iteritems(match.groupdict()):
                    if v:
                        start, end = match.span(k)
                        add_point((start, True, k))
                        add_point((end, False, k))

        def sub_linebreaks(match):
            start = match.start(0)
            line_starts.append(start)
            add_point((start, True, ''))
        self._re_linebreaks.sub(sub_linebreaks, code)

        if highlight_range:
            line_no, start, end = highlight_range
            try:
                line_start = line_starts[line_no - 1]
            except IndexError:
                pass
            else:
                hi_start = line_start + start
                hi_end = line_start + end + 1
                add_point((hi_start, True, highlight_range_style))
                add_point((hi_end, False, highlight_range_style))

        points.sort()
        lines_out = []
        hiline = []
        hiline_append = hiline.append
        style_set = set()
        pos = 0

        points = points[::-1]

        while points:
            start_pos, add_style, style = points.pop()
            new_style_set = style_set.copy()

            if style:
                if add_style:
                    new_style_set.add(style)
                else:
                    new_style_set.discard(style)
                while points:
                    peek_start, peek_add, peek_style = points[-1]
                    if peek_style and peek_start == start_pos:
                        if peek_add:
                            new_style_set.add(peek_style)
                        else:
                            new_style_set.discard(peek_style)
                        points.pop()
                    else:
                        break

            if start_pos > pos:
                hiline_append(_escape_html(code[pos:start_pos]))
            pos = start_pos

            if not style:
                if style_set:
                    hiline.append('</span>')
                lines_out.append(hiline[:])
                del hiline[:]
                if new_style_set:
                    hiline_append('<span class="%s">' % ' '.join(new_style_set))
                style_set = new_style_set
                continue

            if new_style_set != style_set:
                hiline_append('</span>')
                if new_style_set:
                    hiline_append('<span class="%s">' % ' '.join(new_style_set))

            style_set = new_style_set

        if hiline:
            lines_out.append(hiline[:])
            lines_out.append('</span>')

        if end_line is None:
            lines_out = lines_out[start_line - 1:]
        else:
            lines_out = lines_out[start_line - 1:end_line]

        def make_line(l):
            text = ''.join(l)
            if text.strip():
                return text.replace('\n', '')
            else:
                return '\n'
        html_lines = [make_line(line) for line in lines_out]

        if line_numbers:
            html_lines = ['<span class="lineno">{0}</span>{1}'.format(line_no, line)
                          for line_no, line in enumerate(html_lines, (start_line or 0))]

        if highlight_lines is None:
            highlight_lines = ()
        if self.line_anchors:
            linet = '<a name="line{1}"></a><div class="line{0} line-{1}">{2}</div>'
        else:
            linet = '<div class="line{0} line-{1}">{2}</div>'

        html_lines = [(linet.format(' highlight' if line_no in highlight_lines else '', line_no, line))
                      for line_no, line in enumerate(html_lines, (start_line or 0))]

        return "".join(html_lines).replace('\n', '<br>')


class Highlighter(with_metaclass(HighlighterMeta, HighlighterType)):
    pass


class TextHighlighter(Highlighter):
    format = "text"


class XMLHighlighter(Highlighter):
    format = "xml"

    styles = [
        r'(?P<comment><!--.*?-->)|(?P<tag><(?P<xmlns>\w*?:)?(?P<tagname>[\w\-]*)(?P<tagcontent>.*?)(?P<endtagname>/[\w\-:]*?)?>)',
        r'\s\S*?=(?P<attrib>\".*?\")',
        r'(?P<braced>\{.*?\})',
        r'(?P<sub>\$\{.*?\})',
        r'(?P<cdata>\<\!\[CDATA\[.*?\]\]\>)'
    ]


class PythonHighlighter(Highlighter):
    format = "python"

    styles = [
        r'\b(?P<keyword>yield|is|print|raise|pass|and|or|not|return|def|class|import|from|as|for|in|try|except|with|finally|if|else|elif|while)\b',
        r'\b(?P<constant>None|True|False)\b',
        r'\b(?P<builtin>open|file|str|repr|bytes|unicode|int)\b',
        r'\b\((?P<call>.*?)\)',

        r'\b((?:def|class)\s+(?P<def>\w+))',


        #('string', r'(".*?(?<!\\)")'),
        #('string', r'(""".*?""")'),
        #('string', r"('.*?(?<!\\)')"),
        #('string', r"('''.*?''')"),
        r'(?P<self>self)',
        r'\b(?P<number>\d+)',
        r'(?P<operator>\W+)',
        r'\b(?P<operator>or|and|in)\b',
        r'(?P<brace>\(|\)|\[|\]|\{|\})',
        r'@(?P<decorator>[\w\.]*)',

        r'(?P<comment>#.*?)$|(?P<string>(?:""".*?""")|(?:"(?:\\.|.)*?")' + r"|(?:'''.*?''')|(?:'(?:\\.|.)*?'))",

    ]


class HTMLHighlighter(Highlighter):
    format = "html"

    styles = [
        r'(?P<comment><!--.*?-->)|(?P<tag><(?P<xmlns>\w*?:)?(?P<tagname>\w*)(?P<tagcontent>.*?)(?P<endtagname>/\w*?)?>)',
        r'\s\S*?=(?P<attrib>\".*?\")',
    ]


class MoyatemplateHighlighter(Highlighter):
    format = "moyatemplate"

    styles = [
        r'(?P<comment><!--.*?-->)|(?P<tag><(?P<xmlns>\w*?:)?(?P<tagname>\w*)(?P<tagcontent>.*?)(?P<endtagname>/\w*?)?>)',
        r'\s\S*?=(?P<attrib>\".*?\")',

        r'(?P<sub>\$\{.*?\})',
        r'(?P<templatetag>{%.*?%})'
    ]


class RouteHighlighter(Highlighter):
    format = "route"
    styles = [
        r'(?P<special>\{.*?\})'
    ]


class TargetHighlighter(Highlighter):
    line_anchors = False
    format = "target"
    styles = [
        r'(?P<libname>[\w\.\-]+?)(?P<hash>\#)(?P<elementname>[\w\.\-]+)',
    ]


class INIHighlighter(Highlighter):
    line_anchors = False
    format = "ini"
    styles = [
        r'^(?P<key>.*?)=(?P<value>.*?)$',
        r'^(\s+)(?P<value>.*?)$',
        r'^(?P<section>\[.*?\])$',
        r'^(?P<section>\[(?P<sectiontype>.*?)\:(?P<sectionname>.*?)\])$',
        r'^(?P<comment>#.*?)$'
    ]


class JSHighligher(Highlighter):
    format = "js"
    styles = [
        r'(?P<keyword>this|function|var|new|alert)',
        r'\b(?P<constant>null|true|false)\b',
        r'\b(?P<number>\d+)',
        r'(?P<operator>\W+)',
        r'\b(?P<operator>\|\||\&\&|\+|\-|\*|\/)\b',
        r'(?P<brace>\(|\)|\[|\]|\{|\})',
        r'(?P<string>".*?")',
        r'(?P<string>\'.*?\')',
        r'(?P<comment>\/\*.*?\*\/)'
    ]


class Formatter(object):

    def __init__(self, lineno=True):
        self.lineno = lineno


class HTMLFormatter(Formatter):
    pass


class SyntaxFilter(MoyaFilterBase):
    def __moyafilter__(self, context, app, value, params):
        lang = params.pop('lang', None)
        value = dedent(remove_padding(text_type(value.strip('\n'))))
        code = highlight(lang, value, line_numbers=False)
        return HTML('<pre class="moya-console format-{lang}"">{code}</pre>'.format(lang=lang, code=code))


if __name__ == "__main__":
    code = open("console.py", "rb").read()
    code = code.decode('utf-8')
    html = highlight("python", code)
    with timer("highlight", ms=True):
        html = highlight("python", code)
    with open("syntaxtest.html", 'wt') as f:
        f.write(html.encode('utf-8'))



from moya.run import get_callable_from_document

call = get_callable_from_document('mandel.xml', 'mandel')

from moya.tools import timer
with timer("Calculate mandelbrot set", write_file="mandeltimes.csv"):
    call()



from __future__ import unicode_literals
from __future__ import print_function

from threading import Thread
import logging
try:
    import subprocess32 as subprocess
except ImportError:
    import subprocess

log = logging.getLogger('moya.runtime')

from ..context import Context
from ..compat import iteritems
from ..elements.elementbase import Attribute
from ..tags.context import DataSetter, LogicElement
from ..import db
from ..__init__ import pilot


class MoyaThread(Thread):

    def __init__(self, element, app, name, context, data, join_timeout=None):
        super(MoyaThread, self).__init__(name=name)
        self.element = element
        self.app = app
        self.context = context
        self.data = data
        self.join_timeout = join_timeout
        self.libid = element.libid
        self._result = None
        self._error = None

    def __repr__(self):
        return "<thread {} '{}'>".format(self.libid, self.name)

    def run(self):
        archive = self.element.archive
        with pilot.manage(self.context):
            try:
                self._result = archive.call_params(self.element.libid,
                                                   self.context,
                                                   self.app,
                                                   self.data)
            except Exception as e:
                self.context['.console'].obj(self.context, e)
                self._error = e
                dbsessions = self.context['._dbsessions']
                if dbsessions:
                    db.rollback_sessions(self.context)

            else:
                dbsessions = self.context['._dbsessions']
                if dbsessions:
                    db.commit_sessions(self.context)

    def wait(self):
        self.join(self.join_timeout)

    def __moyacontext__(self, context):

        self.join(self.join_timeout)
        if self.join_timeout:
            if self.is_alive():
                self.element.throw('thread.timeout',
                                   'the thread failed to complete within timeout',
                                   thread=self)

        if self._error:
            self.element.throw('thread.fail',
                               'exception occurred in thread',
                               diagnosis="{!r} raised exception '{}'".format(self, self._error),
                               original=self._error,
                               thread=self)
        return self._result


class ThreadElement(DataSetter):
    """
    Run enclosed block in a thread.

    When Moya encounters this tag it executes the enclosed code in a new [url "https://en.wikipedia.org/wiki/Thread_(computing)"]thread[/url]. The calling thread will jump to the end of the [tag]thread[/tag] tag.

    The enclosed code may return a value (with [tag]return[/tag]), which can be retrieved by the calling thread via it's [c]dst[/c] parameter. If the calling thread references the return value before the thread has returned, it will block until the thread returns.

    This tag is useful in a request when you have some long running action to do, and you don't want to delay returning a response. Here's an example:

    [code xml]
    <view libname="view.process" template="slowprocess.html">
        <thread>
            <call macro="slow"/>
        </thread>
    </view>
    [/code]

    This view will render a template and return immediately, while the macro is processing in the background. Note, that the slow macro will have no way of returning anything in the response, but could still send an email or store something in the database to communicate the result to the user.

    """

    class Help:
        synopsis = "run a thread"

    class Meta:
        tag_name = "thread"

    name = Attribute("Name of thread", required=False, default=None)
    scope = Attribute("Use the current scope?", type="boolean", default=False)
    timeout = Attribute("Maximum time to wait for thread to complete", type="timespan", default=None)
    join = Attribute("Join threads before end of request?", type="boolean", default=False)

    def logic(self, context):
        params = self.get_parameters(context)

        thread_context = Context({k: v for k, v in iteritems(context.root) if not k.startswith('_')})
        if '._dbsessions' in context:
            thread_context['._dbsessions'] = db.get_session_map(self.archive)

        data = {}
        if params.scope:
            data.update(context.capture_scope())
        data.update(self.get_let_map(context))

        for k, v in iteritems(data):
            if hasattr(v, '_moyadb'):
                self.throw("thread.not-thread-safe",
                           "thread parameter {} ('{}') may not be passed to a thread".format(context.to_expr(v), k),
                           diagnosis="Database objects are not [i]thread safe[/i], try retrieving the object again inside the thread.")

        moya_thread = MoyaThread(self,
                                 context.get('.app', None),
                                 params.name or self.docname,
                                 thread_context,
                                 data=data,
                                 join_timeout=params.timeout)
        if params.join:
            context.set_new_call('._threads', list).append(moya_thread)
        moya_thread.start()
        self.set_context(context, params.dst, moya_thread)


class WaitOnThreads(LogicElement):
    """
    Wait for threads to complete.

    This tag will will for all [tag]tread[/tag] tags with [c]join[/c] set to [c]yes[/c] to complete.

    """

    class Help:
        synopsis = "wait threads to complete"

    def logic(self, context):
        for thread in context.get('._threads', []):
            thread.wait()
        context.safe_delete('._threads')


class SystemCall(DataSetter):
    """
    Call a system command and get output.

    Commands may be invoked in one of two ways; either with the [c]args[/c] attribute, which should be a list, or with the [c]shell[/c] attribute which should be a string to be passed to the shell. For example, the following two calls are equivalent:

    [code xml]
    <system-call shell="ls -al"/>
    [/code]
    [code xml]
    <system-call args="['ls', '-al']"/>
    [/code]

    The [c]args[/c] form is preferred because [c]shell[/c] is a potential security risk; if you don't escape parameters retrieved from a request it may be possible for a malicious use to execute code on your server.

    If the process returns a non-zero error code, then a [c]system-call.process-error[/c] exception is thrown. The [c]info[/c] value of the exception will contain [c]code[/c] and [c]output[/c].

    """

    class Help:
        synopsis = "run a system command"
        example = """
        <system-call shell="ls -al" console="yes"/>
        """

    class Meta:
        one_of = [('shell', 'args')]

    args = Attribute('call arguments', type="expression", required=False)
    shell = Attribute('shell command', required=False)

    console = Attribute("write output to the console?", type="boolean", default=False, required=False)
    log = Attribute("write output to this log", default=None, required=False)
    output = Attribute("Destination for output", type="reference", default=None, required=False)

    def logic(self, context):
        params = self.get_parameters(context)
        console = context['.console']

        shell = False
        if self.has_parameter('args'):
            try:
                command = list(params.args)
            except:
                self.throw('bad-value.args',
                           "args parameter should be a list")
        else:
            command = params.shell
            shell = True

        def write_log(output, write):
            for line in output.splitlines():
                write(line)

        try:
            output = subprocess.check_output(command,
                                             shell=shell,
                                             stderr=subprocess.STDOUT,
                                             universal_newlines=True)
        except subprocess.CalledProcessError as e:
            return_code = e.returncode
            output = e.output
            if params.console and console:
                console(output)
            if params.log:
                write_log(output, logging.getLogger(params.log).error)
            self.throw('system-call.process-error',
                       'system call returned non-zero code ({})'.format(return_code),
                       code=return_code,
                       output=output)

        if params.console and console:
            console(output)
        if params.log:
            write_log(output, logging.getLogger(params.log).info)
        if params.output:
            context[params.output] = output
        self.set_context(context, params.dst, output)



from __future__ import unicode_literals, print_function

from fs.path import pathjoin, basename
from fs.errors import FSError

import mimetypes
import json


class ReaderError(Exception):
    pass


class UnknownFormat(ReaderError):
    """Attempt to read a format we don't understand"""


class RelativePathError(ReaderError):
    """Can't read from a relative with without an app"""


class DataReader(object):
    """Reads and decodes files of known types"""

    def __init__(self, fs):
        self.fs = fs

    def __repr__(self):
        return "<reader {!r}>".format(self.fs)

    def read(self, path, app=None, mime_type=None):
        """Read a file"""
        if not path.startswith('/'):
            if app is None:
                raise RelativePathError("Can't use relative data paths with an application")
            path = pathjoin(app.data_directory, path)

        filename = basename(path)
        if mime_type is None:
            mime_type, encoding = mimetypes.guess_type(filename)

        _type, sub_type = mime_type.split('/', 1)
        try:
            if mime_type == "text/plain":
                data = self.fs.getcontents(path, mode="rt", encoding="utf-8")
            elif mime_type == "application/json":
                with self.fs.open(path, 'rt', encoding="utf-8") as f:
                    data = json.load(f)
            elif mime_type == "application/octet-stream":
                data = self.fs.getcontents(path, mode="rb")

            elif _type == "text":
                data = self.fs.getcontents(path, mode="rt", encoding="utf-8")

            else:
                raise UnknownFormat("Moya doesn't know how to read file '{}' (in {!r})".format(path, self.fs))
        except FSError as e:
            from .logic import MoyaException
            info = {
                "path": path,
                "mime_type": mime_type
            }
            raise MoyaException("data.read-fail",
                                "unable to read data from {path} ({e})".format(path=path, e=e),
                                diagnosis="check the data exists with **moya fs data --tree**",
                                info=info)

        return data

    def exists(self, path, app):
        """Check if a file exists"""
        if not path.startswith('/'):
            if app is None:
                raise RelativePathError("Can't use relative data paths with an application")
            path = pathjoin(app.data_directory, path)
        try:
            return self.fs.isfile(path)
        except FSError:
            return False


if __name__ == "__main__":

    reader = DataReader(None)
    reader.read('test.bin')



from __future__ import unicode_literals
from __future__ import print_function

import os
from . import syntax

class Frame(object):
    def __init__(self,
                 code,
                 location,
                 lineno,
                 path=None,
                 obj=None,
                 cols=None,
                 one_line=False,
                 libid=None,
                 format="xml",
                 raw_location=None):
        self.code = code
        self._location = location
        self.lineno = lineno
        self.obj = obj
        self.cols = cols
        self.one_line = one_line
        self.format = format
        self.libid = libid
        self._raw_location = raw_location

    def __repr__(self):
        return "<frame '{}'>".format(self.location)

    @property
    def location(self):
        location = self._location
        homedir = os.environ.get('HOME', None)
        if homedir and location.startswith(homedir + '/'):
            location = '~/' + location[len(homedir) + 1:]
        if self.obj:
            return 'File "{}", line {}, in {}'.format(location, self.lineno, self.obj)
        else:
            if self.cols:
                return 'File "{}", line {}, col {}'.format(location, self.lineno, self.cols[0])
            else:
                return 'File "{}"'.format(location)

    @property
    def raw_location(self):
        return self._raw_location or self._location

    @property
    def snippet(self):
        try:
            if not self.code:
                return ''
            if self.one_line:
                return self.code
            lineno = max(0, self.lineno - 5)

            return syntax.highlight(self.format,
                                    self.code,
                                    lineno,
                                    lineno + 10,
                                    highlight_lines=[self.lineno],
                                    highlight_range=[self.lineno, self.cols[0], self.cols[1]] if self.cols else None)
        except Exception as e:
            raise
            from traceback import print_exc
            print_exc(e)



from __future__ import unicode_literals
from __future__ import print_function

from ...command import SubCommand

import os


MOYA_TEMPLATE_XML = """\
<?xml version="1.0" encoding="UTF-8"?>
<moya xmlns="http://moyaproject.com"
    xmlns:moya="http://moyaproject.com"
    xmlns:let="http://moyaproject.com/let"
    xmlns:db="http://moyaproject.com/db"
    xmlns:forms="http://moyaproject.com/forms"
    xmlns:w="http://moyaproject.com/widgets"
    xmlns:html="http://moyaproject.com/html">

    <!-- your content here -->

</moya>
"""


class New(SubCommand):
    """Create a boilerplate moya file"""
    help = "create a new boilerplate moya file"

    def add_arguments(self, parser):
        parser.add_argument(dest='filename', metavar="PATH",
                            help="path / filename for new file")
        parser.add_argument('-f', '-force', dest="force", action="store_true",
                            help="fore overwriting")

    def run(self):
        args = self.args

        if not args.force and os.path.exists(args.filename):
            self.console.error('destination exists, use --force to overwrite')
            return -1

        td = {}
        file_data = MOYA_TEMPLATE_XML.format(td)

        try:
            with open(args.filename, 'wt') as f:
                f.write(file_data)
        except Exception as e:
            self.console.error('unable to write file ({})'.format(e))
            return -1



# PY2 + PY3
from __future__ import unicode_literals
from __future__ import unicode_literals

from .compat import text_type, iteritems
from .containers import OrderedDict

import os
import re

re_section = re.compile(r'\[(.*?)\]', re.UNICODE)


def sub_env(text, _re_env=re.compile(r'\$(\w+)', re.MULTILINE)):
    """Substition renvironment, in $ENV_VARIABLE syntax"""
    get_environ = os.environ.get
    def repl(match):
        return get_environ(match.group(1), match.group(0))
    return _re_env.sub(repl, text)


def parse(inifile, sections=None, section_class=OrderedDict, _sub_env=sub_env):
    """Parse an ini file in to nested dictionaries"""
    if hasattr(inifile, "read"):
        ini = inifile.read()
    else:
        ini = inifile
    if not isinstance(ini, text_type):
        ini = ini.decode('utf-8')
    inilines = ini.splitlines()

    if sections is None:
        sections = section_class()
    current_section = ''
    current_section_data = section_class()
    current_key = None
    current_value = ''

    section_match = re_section.match

    for line in inilines:
        if line.startswith('#'):
            continue
        if not line.strip():
            current_key = None
            continue
        match = section_match(line)
        if match:
            sections[current_section] = current_section_data
            current_section_data = section_class()
            current_section = match.group(1)
        elif line[0] in ' \t':
            if current_key is not None:
                current_value += '\n' + line.strip()
                current_section_data[current_key] = current_value
        elif '=' in line:
            key, value = line.split('=', 1)
            key = key.rstrip()
            value = _sub_env(value.lstrip()).lstrip()
            current_key = key
            current_value = value
            current_section_data[key] = value
        else:
            current_section_data[line.strip()] = ''

    sections[current_section] = current_section_data
    return sections


def write(settings, comments=None):
    """Write an ini file from nested dictionaries"""
    if comments is None:
        comments = []
    if isinstance(comments, text_type):
        comments = comments.splitlines()
    lines = ["# " + comment for comment in comments]

    def write_section(name, section):
        if name:
            lines.append("[{}]".format(name))
        for k, v in iteritems(section):
            v = "\n    ".join(v.split('\n'))
            lines.append("{k} = {v}".format(k=k, v=v))
        lines.append('')

    if '' in settings:
        write_section('', settings[''])

    for name, section in iteritems(settings):
        if name:
            write_section(name, section)

    return '\n'.join(lines)



if __name__ == "__main__":

    ini = """# -------------------------------------------------------------
# Filesystems
# -------------------------------------------------------------

foo=bar

[fs:project]
location = ./
multiline = long
    line

[templateengine:jinja2]
cachefs = jinja2cache

[templates:default]
location = ./templates
priority = 10

[app:blog]
name = Will's blog
value = 3

noequal=

"""

    settings = parse(ini)

    settings["new"] = {"foo": "bar\nbaz"}

    print(write(settings, comments=["Re-written ini file", "comments"]))



from __future__ import unicode_literals
from ..elements.elementbase import ElementBase, LogicElement, Attribute
from ..tags.context import DataSetter
from ..tools import asint
from .. import logic
from .. import errors
from ..elements.attributetypes import *
from ..contextenum import ContextEnum
from ..compat import iteritems, text_type, reload
from .. import tools
from .. import timezone

from fs.path import pathjoin
from fs.opener import fsopendir
from fs.errors import FSError

from os.path import dirname, abspath
import sys
import textwrap
from time import time

import pytz

import logging
log = logging.getLogger('moya.runtime')
startup_log = logging.getLogger('moya.startup')
runtime_log = logging.getLogger('moya.runtime')


class Moya(ElementBase):
    """This is the root element for Moya files."""
    version = Attribute("""version number of this Moya file (currently ignored, if supplied use "1.0")""", type="version", required=False,)

    class Help:
        synopsis = """begin a Moya file"""
        example = """
        <moya xmlns="http://moyaproject.com">
            <!-- Ready for take off -->
        </moya>

        """


class ConfigElement(ElementBase):

    class Help:
        undocumented = True

    def execute(self, archive, context, fs):
        pass


class Breakpoint(LogicElement):
    """Stops the execute of Moya code and drops in to the debugger."""

    class Help:
        synopsis = """drop in to the debugger"""
        example = """

        <!-- drop in to the debugger -->
        <breakpoint/>

        <for src="-100..100" dst="count">
            <!-- Conditional breakpoint with 'if' attribute -->
            <breakpoint if="count==0"/>
            <!-- This throws an exception when count is 0 -->
            <echo>${1/count}<echo>
        </for>

        """

    def logic(self, context):
        raise logic.DebugBreak(self)


class Import(LogicElement):
    """Import a library. Importing reads the XML associated with a library and makes it available to be installed as an application. This tag must appear within a [tag]server[/tag] tag.

    Libraries can be installed either from a Python module, or directly from a path. For example, this installs a library from a Python module:

    [code xml]
    <import py="moya.libs.auth"/>
    [/code]

    And this installs a library from a a relative path:

    [code xml]
    <import location="./local/sushishop/" />
    [/code]

    The [c]priority[/c] attribute is used when two element have the same element reference. This is typically used to override an element in another library. For example, let's say we have the following [tag]macro[/tag] in a library called [c]sushifinder.shop[/c], which calculate tax on a shopping cart:

    [code xml]
    <macro libname="macro.get_tax">
        <!-- calculate tax for an order-->
    </macro>
    [/code]

    We could replace this by defining the following in another library:

    [code xml]
    <macro libname="sushifinder.shop#macro.get_tax">
        <!-- custom tax calculator -->
    </macro>
    [/code]

    Note the use of a full element reference (including library name) in the [c]libname[/c] attribute. This tells Moya that the macro should go in a library other than the one it was declared in. If the code above is in a library with a higher priority then it will replace the macro in the [c]sushifinder.shop[/c] library.

    The [c]templatepriority[/c] is used when there are conflicting template paths. The template from the library with the highest priority wins. This value takes precedence over the [c]priority[/c] defined in [link library#lib-section]lib.ini[/link].


    """

    class Help:
        synopsis = """import a library"""

    name = Attribute("Name of the library")
    location = Attribute("A path to a Moya library")
    py = Attribute("Python import, e.g. widgets.moya.widgetapp")
    priority = Attribute("Priority for elements", type="integer", required=False, default=None)
    templatepriority = Attribute("Priority for templates", type="integer", required=False, default=None)

    def logic(self, context):
        start = time()
        (name,
         _location,
         py,
         priority,
         template_priority) = self.get_parameters(context,
                                                  'name',
                                                  'location',
                                                  'py',
                                                  'priority',
                                                  'templatepriority')
        if template_priority is None:
            template_priority = priority
        archive = self.document.archive
        absolute = False
        if _location is not None:
            location = _location
        else:
            if py in sys.modules:
                reload(sys.modules[py])
            try:
                __import__(py)
            except ImportError as e:
                raise errors.ElementError("unable to import Python module '{}'".format(py),
                                          element=self,
                                          diagnosis=text_type(e))
            module = sys.modules[py]
            location = dirname(abspath(module.__file__))
            absolute = True

        try:

            if '::/' in location:
                import_fs = fsopendir(location)
            else:
                if absolute:
                    import_fs = fsopendir(location)
                else:
                    project_fs = context['fs']
                    try:
                        if project_fs.hassyspath('/'):
                            project_path = project_fs.getsyspath('/')
                            import_path = pathjoin(project_path, location)
                            import_fs = fsopendir(import_path)
                        else:
                            import_fs = context['fs'].opendir(location)
                    except FSError as e:
                        self.throw("import.fail",
                                   "unable to import library from {}".format(location),
                                   diagnosis=text_type(e))
            lib = archive.load_library(import_fs,
                                       priority=priority,
                                       template_priority=template_priority,
                                       long_name=name,
                                       rebuild=context.root.get('_rebuild', False))
            if lib.failed_documents:
                if _location is not None:
                    msg = "Failed to load library '{}' from location '{}'"
                    raise errors.StartupFailedError(msg.format(name or lib.long_name, _location))
                elif py:
                    msg = "Failed to load library '{}' from Python module '{}'"
                    raise errors.StartupFailedError(msg.format(name or lib.long_name, py))
                else:
                    raise errors.StartupFailedError("Failed to load library '{}'".format(name or lib.long_name))
            startup_log.debug("%s imported %.1fms", lib, (time() - start) * 1000.0)
            if lib.priority:
                startup_log.debug("%s priority is %s", lib, lib.priority)
            if lib.template_priority:
                startup_log.debug("%s template priority is %s", lib, lib.template_priority)
        except Exception as e:
            if not self.archive.test_build:
                raise


class Install(LogicElement):
    """
    Installs an [i]application[/i]. An application is effectively an [i]instance[/i] of a library, with its own settings and serving content from unique URL(s). An [tag]install[/tag] tag should appear within a [tag]server[/tag] tag. A library must first be [i]imported[/i] (with [tag]import[/tag]) prior to installing.

    """

    class Help:
        synopsis = """create an application from a library"""
        example = """
        <install name="auth" lib="moya.auth" mount="/auth/"/>
        """

    lib = Attribute("Library long name (e.g. moya.auth)", required=True)
    name = Attribute("Name of the application (must not contain a dot), e.g. \"auth\"", required=True)
    mount = Attribute("URL component to mount, e.g. \"auth\"")
    mountpoint = Attribute("Name of the <mountpoint> tag", required=False, default="main")
    urlpriority = Attribute("Priority for URLs in mountpoint", type="integer", required=False, default=0)

    def logic(self, context):
        params = self.get_parameters(context)
        archive = self.document.archive
        try:
            self.archive.build_libs()
        except Exception as e:
            if not self.archive.test_build:
                raise

        try:
            try:
                app = archive.create_app(params.name, params.lib)
            except errors.ArchiveError as e:
                raise errors.ElementError(text_type(e))
            if app.lib.failed_documents:
                raise errors.StartupFailedError("Unable to import lib '%s'" % params.lib)

            if params.mount:
                server = self.get_ancestor('server')
                try:
                    mountpoint = app.lib.get_element_by_type_and_attribute("mountpoint", "name", params.mountpoint)
                except errors.ElementNotFoundError:
                    return
                    raise errors.StartupFailedError("No mountpoint called '{0}' in {1}".format(params.mountpoint, app.lib))
                app.mounts.append((params.mountpoint, params.mount))
                server.urlmapper.mount(params.mount,
                                       mountpoint.urlmapper,
                                       defaults={'app': app.name},
                                       name=params.name,
                                       priority=params.urlpriority)

                for stage, urlmapper in iteritems(server.middleware):
                    urlmapper.mount(params.mount,
                                    mountpoint.middleware[stage],
                                    defaults={'app': app.name},
                                    name=params.name,
                                    priority=params.urlpriority)
                startup_log.debug("%s installed, mounted on %s", app, tools.normalize_url_path(params.mount))
            else:
                startup_log.debug("%s installed", app)
        except:
            if not self.archive.test_build:
                raise


class Log(LogicElement):
    """Write text to the logs. Every line of text can have a [i]level[/i] which indicates how significant the message is. Log levels can be filtered, so you only see relevant messages.

    You can control which log messages are written via [c]logging.ini[/c].
    """

    class Help:
        synopsis = """write information to the log"""
        example = """

        <log level="debug">This may help you track down errors</log>
        <log>Something worth knowing happened</log> <!-- default log level is "info" -->
        <log level="warning">Pay attention, this may be significant</log>
        <log level="error">Something quite alarming happened</log>
        <log level="fatal">Absolute disaster</log>

        """

    _levels = {
        "debug": 10,
        "info": 20,
        "warn": 30,
        "warning": 30,
        "error": 40,
        "fatal": 50
    }
    _default_level = "info"

    level = Attribute('''Logging level''',
                      required=False,
                      default=None,
                      choices=_levels.keys())
    logger = Attribute("Logger to write to",
                       default=None,
                       required=False)

    def logic(self, context):
        text = textwrap.dedent(context.sub(self.text))
        _level, _logger = self.get_parameters(context, 'level', 'logger')
        _level = _level or self._default_level
        if _level.isdigit():
            level = int(_level)
        else:
            level = self._levels.get(_level, logging.INFO)
        if _logger:
            log = logging.getLogger(_logger)
        else:
            app_name = context.get('.app.name', None)
            if app_name is None:
                log = runtime_log
            else:
                log = logging.getLogger('moya.app.{}'.format(app_name))
        for line in text.splitlines():
            if line:
                log.log(level, line)


class LogDebug(Log):
    """See [tag]log[/tag]"""
    _default_level = "debug"

    class Help:
        synopsis = "write information to the log"


class LogInfo(Log):
    """See [tag]log[/tag]"""
    _default_level = "info"

    class Help:
        synopsis = "write information to the log"


class LogWarn(Log):
    """See [tag]log[/tag]"""
    _default_level = "warn"

    class Help:
        synopsis = "write information to the log"


class LogError(Log):
    """See [tag]log[/tag]"""
    _default_level = "error"

    class Help:
        synopsis = "write information to the log"


class LogFatal(Log):
    """See [tag]log[/tag]"""
    _default_level = "fatal"

    class Help:
        synopsis = "write information to the log"


class Choices(ElementBase):
    """
    Define choices, for use by select fields.

    Choices are a list of values and labels, that might be used by a select control. You can retrieve the data defined by this tag with [tag]get-choices[/tag].

    """

    class Help:
        synopsis = "define choices"
        example = """
            <choices libname="choices.markup">
                <choice value="text" label="Plain text"/>
                <choice value="bbcode" label="BBCode markup"/>
                <choice value="html" label="HTML"/>
            </choices>
            <get-choices choices="#choices.markup" dst="choices" />
        """

    def finalize(self, context):
        self.choices = choices = []
        append = choices.append
        for choice in self.get_children('choice'):
            value, label = choice.get_parameters(context, 'value', 'label')
            if not label:
                label = context.sub(choice.text)
            append((value, label))


class Choice(ElementBase):
    """
    Define a choice in a [tag]choices[/tag] tag.

    """

    class Help:
        synopis = "define a choice"

    value = Attribute("Choice value")
    label = Attribute("Label to display")


class GetChoices(DataSetter):
    """
    Get choice data from a [tag]choices[/tag] tag

    """
    choices = Attribute("Choices element", type="elementref", required="yes")
    _from = Attribute("Application", type="application", required=False, default=None)


    def get_value(self, context):
        app, choices = self.get_parameters(context, 'from', 'choices')
        app = app or context['.app']
        _app, choices_el = app.get_element(choices)
        return choices_el.choices


class Enum(ElementBase):
    """
    Define and [i]enumeration[/i] object. An enumeration is a collection of text identifiers with an integer value. This tag should contain [tag]value[/tag] tags that define the values.

    You can retrieve an enumeration object with the [tag]get-enum[/tag] tag.

    """
    name = Attribute('Identifier for enum', required=False, default=None)
    start = Attribute("Starting ID if not specified in <value>", type="integer", default=1)

    class Help:
        synopsis = """map numbers on to identifiers"""
        example = """
        <enum libname="enum.jsonrpc.errors">
            <value id="1" name="not_logged_in" description="You must be logged in to do that" />
            <value name="invalid_score" description="Score must be -1, 0 or +1"/>
            <value name="unknown_link" description="Link object was not found"/>
        </enum>
        """

    def post_build(self, context):
        params = self.get_parameters(context)
        start = params.start
        name = params.name or self.libname

        enum = ContextEnum(name, start=start)
        self.archive.add_enum(self.libid, enum)
        #startup_log.debug("%s created", enum)


class Value(ElementBase):
    """Defines a single value in an enumeration object. Should be contained within an [tag]enum[/tag]."""

    class Help:
        synopsis = """a value in an enumeration"""

    id = Attribute("Enumeration ID", type="integer", default=None)
    name = Attribute("Value name", required=True)
    description = Attribute("Description of enumeration value", default='')
    group = Attribute("Group name", default='')

    def finalize(self, context):
        enum_parent = self.get_ancestor('enum')
        params = self.get_parameters(context)
        description = params.description
        if not description:
            description = context.sub(self.text)
        enum = self.archive.get_enum(enum_parent.libid)
        if description:
            description = description.strip()
        enum.add_value(params.name,
                       enum_id=params.id,
                       description=description,
                       group=params.group)


class GetEnum(DataSetter):
    """Get an enumeration object (see [tag]enum[/tag])."""

    class Help:
        synopsis = "get an enumeration"
        example = """
        <get-enum enum="sociallinks#enum.jsonrpc.errors" dst="errors" />
        <return value="errors.unknown_link" />

        """

    enum = Attribute("enumeration ref", type="elementref")
    _from = Attribute("from application", type="application", default=None, evaldefault=True)

    def logic(self, context):
        params = self.get_parameters(context)
        app = self.get_app(context)
        app, el = app.get_element(params.enum)
        enum = self.archive.get_enum(el.libid)
        self.set_context(context, params.dst, enum)

        #enum = self.get_element(self.enum(context))
        #self.set_context(context, dst, enum)


class GetTimezones(DataSetter):
    """Get a list of common timezones."""

    class Help:
        synopsis = """get common timezones"""
        example = """
        <get-timezones dst="timezones" />
        <echo>Timezones: ${commalist:timezones}</echo>

        """

    def logic(self, context):
        timezones = pytz.common_timezones[:]
        self.set_context(context, dst, timezones)


class GetTimezoneGroups(DataSetter):
    """Get timezone information, for use in select controls."""

    class Help:
        synopsis = "get timezone choices"

    def get_value(self, context):
        return timezone.get_common_timezones_groups()


class Handle(LogicElement):
    """
    Handles a [i]signal[/i]. A signal is a way of broadcasting to other parts of the project that a particular event has occurred.

    See [doc signals] for more information.

    """

    class Help:
        synopsis = """respond to a signal"""
        example = """

        <!-- Moya fires a number of signals itself, for various system events -->
        <handle signal="sys.startup">
            <log>This code executes on startup, prior to any requests</log>
        </handle>
        <!-- Handle a custom signal -->
        <handle signal="mordor.arrived">
            <log>${hobbit} has arrived in Mordor!</log>
        </handle>

        """

    signal = Attribute("Signal name to handle. Multiple names may be specified to handle more than one signal.""", type="commalist", required=True)
    sender = Attribute("Only handle the signal(s) if sent from this element(s).", type="commalist", default=None)

    def lib_finalize(self, context):
        sender, signals = self.get_parameters(context, 'sender', 'signal')
        senders = []
        if sender:
            for _sender in sender:
                sender = self.document.qualify_element_ref(_sender, lib=self.lib)
                app, element = self.get_element(sender)
                senders.append(element.libid)
        for signal in signals:
            if senders:
                for sender in senders:
                    self.archive.signals.add_handler(signal, self.libid, sender)
            else:
                self.archive.signals.add_handler(signal, self.libid, None)


class Fire(LogicElement):
    """
    Fire (broadcast) a signal. Additional data may be provided to the signal handlers, by setting values in the [i]let map[/i]. Signals may be [i]handlers[/i] with the [tag]handle[/tag] tag.

    Moya will catch and log any exceptions raised by the signal handler(s).

    """
    signal = Attribute("Signal name, should be in a dotted notation. Names with no dots are reserved by Moya.", required=True)
    sender = Attribute("Optional element associated with the signal.", required=False, type="elementref", default=None)
    _from = Attribute("Application", type="application", required=False, default=None)

    class Help:
        synopsis = """fire a signal"""
        example = """
        <fire signal="mordor.arrived" let:hobbit="Frodo"/>
        """

    class Meta:
        trap_exceptions = True
        #is_call = True

    def logic(self, context):
        app = self.get_app(context)
        signal = self.signal(context)
        sender = self.sender(context)
        data = self.get_let_map(context)
        self.archive.fire(context,
                          signal,
                          app=app,
                          sender=sender,
                          data=data)



from __future__ import unicode_literals
from __future__ import absolute_import

from .. import namespaces
from ..elements import Attribute
from ..elements.elementbase import LogicElement
from ..tags.context import DataSetter
from ..compat import text_type

from fs.errors import FSError
from fs.path import pathjoin, basename, dirname

import hashlib
import logging

log = logging.getLogger('moya.fs')


class SetContents(LogicElement):
    """Set the contents of a file"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "write data to a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Destination path", required=True)
    contents = Attribute("File contents", type="expression", required=True, missing=False)

    def logic(self, context):
        params = self.get_parameters(context)
        params.contents
        if self.has_parameter('fsobj'):
            dst_fs = params.fsobj
        else:
            dst_fs = self.archive.lookup_filesystem(self, params.fs)
        try:
            dst_fs.makedir(dirname(params.path), recursive=True, allow_recreate=True)
            dst_fs.setcontents(params.path, params.contents)
        except Exception as e:
            self.throw("fs.set-contents.fail", "unable to set file contents ({})".format(e))
        log.debug("setcontents '%s'", params.path)


class RemoveFile(LogicElement):
    """Delete a file from a filesystem"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "delete a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Destination path", required=True)
    ifexists = Attribute("Only remove if the file exists?", type="boolean", required=False)

    def logic(self, context):
        params = self.get_parameters(context)
        if self.has_parameter('fsobj'):
            dst_fs = params.fsobj
        else:
            dst_fs = self.archive.lookup_filesystem(self, params.fs)
        if params.ifexists and dst_fs.isfile(params.path):
            return
        try:
            dst_fs.remove(params.path)
        except Exception as e:
            self.throw("fs.remove-file.fail",
                       "unable to remove '{}' ({})".format(params.path, e))
        log.debug("removed '%s'", params.path)


class GetSyspath(DataSetter):
    """
    Get a system path for a path in a filesystem.

    A system path (or 'syspath') is a path that maps to the file on the OS filesystem. Not all filesystems can generate syspaths. If Moya is unable to generate a syspath it will throw a [c]get-syspath.no-syspath[/c] exception.

    """
    xmlns = namespaces.fs

    class Help:
        synopsis = "get a system path"
        example = """
        <fs:get-syspath fs="templates" path="index.html" dst="index_template_path"/>
        """

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Destination path", required=True)

    def logic(self, context):
        params = self.get_parameters(context)

        if self.has_parameter('fsobj'):
            dst_fs = params.fsobj
        else:
            dst_fs = self.archive.lookup_filesystem(self, params.fs)

        try:
            syspath = dst_fs.getsyspath(params.path)
        except:
            self.throw('fs.get-syspath.no-syspath',
                       "{!r} can not generate a syspath for '{}'".format(dst_fs, params.path))
        self.set_context(context, self.dst(context), syspath)


class GetMD5(DataSetter):
    """Get the MD5 of a file"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "get the md5 of a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Path of file", required=True)

    def get_value(self, context):
        params = self.get_parameters(context)
        if self.has_parameter('fsobj'):
            fs = params.fsobj
        else:
            fs = self.archive.lookup_filesystem(self, params.fs)

        m = hashlib.md5()
        try:
            with fs.open(params.path, 'rb') as f:
                while 1:
                    chunk = f.read(16384)
                    if not chunk:
                        break
                    m.update(chunk)
        except FSError:
            self.throw("fs.get-md5.fail", "unable to read file '{}'".format(params.path))
        else:
            return m.hexdigest()


class GetInfo(DataSetter):
    xmlns = namespaces.fs

    class Help:
        synopsis = "get an info object for a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Path of file", type="expression", required=True)

    def get_value(self, context):
        params = self.get_parameters(context)
        if self.has_parameter('fsobj'):
            fs = params.fsobj
        else:
            fs = self.archive.lookup_filesystem(self, params.fs)

        try:
            info = fs.getinfo(params.path)
        except FSError:
            self.throw('fs.get-info.fail', "unable to get info for path '{}'".format(params.path))
        else:
            return info


class GetSize(DataSetter):
    xmlns = namespaces.fs

    class Help:
        synopsis = "get the size of a file"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    fs = Attribute("Filesystem name", required=False, default=None)
    path = Attribute("Path of file", type="expression", required=True)

    def get_value(self, context):
        params = self.get_parameters(context)
        path = text_type(params.path)
        if self.has_parameter('fsobj'):
            fs = params.fsobj
        else:
            fs = self.archive.lookup_filesystem(self, params.fs)

        try:
            info = fs.getsize(path)
        except FSError:
            self.throw('fs.get-size.fail', "unable to get info for path '{}'".format(params.path))
        else:
            return info


class Walk(DataSetter):
    """Get a list of files"""
    xmlns = namespaces.fs

    class Help:
        synopsis = "go through the files in a directory"

    fsobj = Attribute("Filesystem object", required=False, default=None)
    path = Attribute("Path to walk", required=False, default="/")
    fs = Attribute("Filesystem name", required=False, default=None)
    dirs = Attribute("Filter directories (function should reference 'name' and return a boolean)", type="function", default=None)
    files = Attribute("Filter files (function should reference 'name' and return a boolean)", type="function", default=None)
    search = Attribute("Search method", default="breadth", choices=["breadth", "depth"])
    dst = Attribute("Destination", required=True, type="reference")

    def logic(self, context):
        params = self.get_parameters(context)
        if self.has_parameter('fsobj'):
            walk_fs = params.fsobj
        else:
            walk_fs = self.archive.get_filesystem(params.fs)

        wildcard = lambda name: params.files(context, name=basename(name)) if self.has_parameter('files') else lambda name: True
        dir_wildcard = lambda name: params.dirs(context, name=basename(name)) if self.has_parameter('dirs') else lambda name: True

        paths = []
        add_path = paths.append

        for dirname, dir_paths in walk_fs.walk(path=params.path,
                                               search=params.search,
                                               wildcard=wildcard,
                                               dir_wildcard=dir_wildcard,
                                               ignore_errors=True):

            for path in dir_paths:
                add_path(pathjoin(dirname, path))
        self.set_context(context, params.dst, paths)



from __future__ import absolute_import
import json


class _MoyaEncoder(json.JSONEncoder):
    """A customer encoder for Moya objects"""
    def default(self, obj):
        if hasattr(obj, '__moyajson__'):
            return obj.__moyajson__()
        return super(_MoyaEncoder, self).default(obj)


def dumps(obj, *args, **kwargs):
    """Allows objects to define how they are serialized to JSON,
    if an object contains a '__moyajson__' method, that will be called and the
    result will be encoded rather than the instance"""
    return json.dumps(obj, cls=_MoyaEncoder, *args, **kwargs)


loads = json.loads



from __future__ import unicode_literals
from __future__ import print_function

from moya import tools
from moya.context.missing import Missing
from moya.compat import text_type
from moya.elements.elementbase import ReturnContainer

import pytz
import datetime
import unittest
import io
import os.path


class TestTools(unittest.TestCase):

    def test_extract_namespace(self):
        tests = [
            ("{http://moyaproject.com}test1", ("http://moyaproject.com", "test1")),
            ("test2", ("http://moyaproject.com", "test2")),
            ("{http://moyaproject.com/db}query", ("http://moyaproject.com/db", "query"))
        ]
        for test, result in tests:
            self.assertEqual(tools.extract_namespace(test), result)

    def test_asint(self):
        assert tools.asint('5') == 5
        assert tools.asint('-5') == -5
        assert tools.asint('foo', 3) == 3

    def test_match_exception(self):

        tests = [
            ("*", 'anything', True),
            ("foo", "foo", True),
            ("foo.bar", "foo.bar", True),
            ("foo.*", "foo.bar", True),
            ("foo.*", "foo.bar.baz", True),
            ("bar", "foo", False),
            ("foo.bar.*", "foo.baz.egg", False)
        ]

        for m, exc, result in tests:
            print(exc, m, result)
            assert tools.match_exception(exc, m) == result

    def test_md5_hexdigest(self):
        assert tools.md5_hexdigest('foo') == 'acbd18db4cc2f85cedef654fccc4a4d8'

    def test_check_missing(self):
        tools.check_missing({'foo': 'bar'})
        try:
            tools.check_missing({'foo': Missing('bar')})
        except ValueError:
            pass
        else:
            assert False

    def test_timer(self):
        with tools.timer('foo'):
            pass
        with tools.timer('foo', ms=True):
            pass
        with tools.timer('foo', write_file="/tmp/__timertest__"):
            pass

    def test_parse_timedelta(self):
        assert tools.parse_timedelta('10') == 10
        assert tools.parse_timedelta('10s') == 10 * 1000
        assert tools.parse_timedelta('1m') == 60 * 1000
        try:
            tools.parse_timedelta('agfdwrg')
        except ValueError:
            assert True
        else:
            assert False

    def test_get_moya_dir(self):
        moya_dir = os.path.join(os.path.dirname(__file__), 'moyadir')
        path = os.path.join(moya_dir, 'foo')
        assert tools.get_moya_dir(path) == moya_dir
        try:
            tools.get_moya_dir()
        except ValueError:
            assert True
        else:
            assert False
        try:
            tools.get_moya_dir('/')
        except ValueError:
            assert True
        else:
            assert False

    def test_is_moya_dir(self):
        moya_dir = os.path.join(os.path.dirname(__file__), 'moyadir')
        path = os.path.join(moya_dir, 'foo')
        assert not tools.is_moya_dir(path)
        assert tools.is_moya_dir(moya_dir)
        assert not tools.is_moya_dir('/')
        assert not tools.is_moya_dir()

    def test_file_chunker(self):
        text = b"Hello, World"
        f = io.BytesIO(text)
        chunks = list(tools.file_chunker(f, 1))
        assert chunks == [b'H', b'e', b'l', b'l', b'o', b',', b' ', b'W', b'o', b'r', b'l', b'd']

        f = io.BytesIO(text)
        chunks = list(tools.file_chunker(f, 2))
        assert chunks == [b'He', b'll', b'o,', b' W', b'or', b'ld']

        f = io.BytesIO(text)
        chunks = list(tools.file_chunker(f, 256))
        assert chunks == [text]

    def test_make_id(self):
        assert tools.make_id() != tools.make_id()

    def test_datetime_to_epoch(self):
        assert tools.datetime_to_epoch(100) == 100
        epoch_start = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=pytz.UTC)
        self.assertEqual(tools.datetime_to_epoch(epoch_start), 0)

    def test_split_commas(self):
        assert tools.split_commas('foo, bar') == ['foo', 'bar']

    def test_summarize_text(self):
        assert tools.summarize_text(None) == ''
        assert tools.summarize_text('hello') == 'hello'
        assert tools.summarize_text('hello, world', max_length=5) == 'hello[...]'

    def test_get_return(self):
        assert tools.get_return(None) == {}
        assert tools.get_return(100) == 100
        ret = ReturnContainer('foo')
        assert tools.get_return(ret) == 'foo'

    def test_as_dict(self):
        assert tools.as_dict({'foo': 'bar'}) == {'foo': 'bar'}

        class D(object):
            def items(self):
                return [('foo', 'bar')]

            def iteritems(self):
                return iter(self.items())
        d = D()
        assert tools.as_dict(d) == {'foo': 'bar'}

    def test_quote(self):
        assert tools.quote('hello') == '"hello"'

    def test_squote(self):
        assert tools.squote('hello') == "'hello'"

    def test_textual_list(self):
        assert tools.textual_list(["foo", "bar"]) == "'foo' or 'bar'"
        assert tools.textual_list(["foo", "bar", "baz"]) == "'foo', 'bar' or 'baz'"
        assert tools.textual_list(["foo"]) == "'foo'"
        assert tools.textual_list([], empty='nadda') == "nadda"

    def test_moya_update(self):
        d = {}
        tools.moya_update(d, {'foo': 'bar'})
        self.assertEqual(d, {'foo': 'bar'})

    def test_url_join(self):
        assert tools.url_join('http://moyaproject.com/', '/foo/') == 'http://moyaproject.com/foo/'

    def test_remove_padding(self):
        assert tools.remove_padding('    ') == ''
        assert tools.remove_padding('') == ''

        assert tools.remove_padding('  hello  ') == '  hello  '
        assert tools.remove_padding('\n\nhello\n\n') == 'hello'

        assert tools.remove_padding('\n\nhello\nworld\n\n') == 'hello\nworld'

    def test_unique(self):
        assert tools.unique([]) == []
        assert tools.unique(['foo']) == ['foo']
        assert tools.unique(['foo', 'bar']) == ['foo', 'bar']
        assert tools.unique(['foo', 'bar', 'bar', 'bar', 'baz']) == ['foo', 'bar', 'baz']
        assert tools.unique(5) == []

    def test_format_element_type(self):
        assert tools.format_element_type(('foo', 'bar')) == "{foo}bar"
        assert tools.format_element_type('foo') == "foo"

    # def test_get_ids(self):
    #     class O(object):
    #         id = 1
    #     assert tools.get_ids([O(), None]) == [1]

    def test_multi_replace(self):
        replacer = tools.MultiReplace({'foo': 'bar', 'baz': 'egg'})
        assert replacer('foo baz foo ok') == "bar egg bar ok"

    def test_dummy_lock(self):
        with tools.DummyLock() as _lock:
            pass

    def test_make_cache_key(self):
        assert tools.make_cache_key(['foo', 'bar']) == 'foo.bar'
        assert tools.make_cache_key(['foo', 'bar', [1, 2]]) == 'foo.bar.1-2'
        assert tools.make_cache_key(['foo', 'bar', [1, 2], {'foo': 'bar'}]) == 'foo.bar.1-2.foo_bar'

    def test_nearers_word(self):
        assert tools.nearest_word('floo', ['foo', 'bar']) == 'foo'

    def test_show_tb(self):
        @tools.show_tb
        def test():
            raise Exception('everything is fine')
        try:
            test()
        except:
            assert True
        else:
            assert False

    def test_normalize_url_path(self):
        assert tools.normalize_url_path('') == '/'
        assert tools.normalize_url_path('foo') == '/foo/'
        assert tools.normalize_url_path('foo/bar') == '/foo/bar/'

    def test_lazystr(self):
        s = tools.lazystr(lambda: 'foo')
        assert text_type(s) == 'foo'
        s = tools.lazystr(lambda: 'foo')
        assert len(s) == 3
        s = tools.lazystr(lambda: 'foo')
        assert s.upper() == 'FOO'



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import

from ..compat import with_metaclass
from ..loggingconf import init_logging_fs

import logging
log = logging.getLogger('moya.startup')


class SubCommandMeta(type):
    registry = {}

    def __new__(cls, name, base, attrs):
        new_class = type.__new__(cls, name, base, attrs)
        if name != "SubCommand":
            cls.registry[name.lower()] = new_class
        return new_class


class SubCommandType(object):
    help = ''
    description = ''

    def __init__(self, command):
        self.command = command
        self.console = self.command.console

    def add_arguments(self, parser):
        pass

    def debug(self, text):
        return self.command.debug(text)

    def error(self, text):
        return self.command.error(text)

    def run(self):
        location = self.location
        # TODO: make logging work from a FS object
        init_logging_fs(self.location_fs, self.args.logging)
        log.debug('project found in "%s"', location)

    # def init_logging(self, location):
    #     init_logging(logging)

    @property
    def location(self):
        return self.command.location

    @property
    def location_fs(self):
        return self.command.location_fs

    def get_settings(self):
        return self.command.get_settings()


class SubCommand(with_metaclass(SubCommandMeta, SubCommandType)):
    pass



from __future__ import unicode_literals, print_function, absolute_import

exposed_elements = {}
exposed_filters = {}


def macro(libname):
    def deco(f):
        exposed_elements[libname] = f
        return f
    return deco


class PyFilter(object):

    def __init__(self, callable):
        self.callable = callable

    def __repr__(self):
        return "<pyfilter '{}'>".format(self.callable.func_name)

    def __moyafilter__(self, context, app, value, params):
        return self.callable(value)


def filter(filtername):
    def deco(f):
        exposed_filters[filtername] = PyFilter(f)
        return f
    return deco



"""
Manage Moya package dependencies

"""

from __future__ import unicode_literals
from __future__ import print_function

from moya import jsonrpc
from moya import settings
from moya import versioning
from collections import OrderedDict

import io
import requests


class DependencyError(Exception):
    pass


def gather_dependencies(rpc, app_name, mount, package, console, no_deps=False):

    visited = set()
    package_stack = [(app_name, mount, package)]

    requirements = OrderedDict()
    dependancy = False

    while package_stack:
        app_name, mount, package = package_stack.pop()
        if package in visited:
            continue
        visited.add(package)

        package_select = rpc.call('package.select', package=package)

        if dependancy and package_select['version'] is None:
            console.text("dependency '{}'' has no installation candidate".format(package), fg="red")
            continue

        name_version = "{} {}".format(package_select['name'], package_select['version'])
        if dependancy:
            console.text("selected {} (dependency)".format(name_version), italic=True)
        else:
            console.text("selected {}".format(name_version), italic=True)
        name = package_select['name']

        requirements[name] = (app_name, mount, package_select)

        if no_deps:
            break

        lib_ini_url = package_select['download'] + '/lib.ini'
        lib_settings_response = requests.get(lib_ini_url, verify=False)
        lib_settings = settings.SettingsContainer.read_from_file(io.StringIO(lib_settings_response.text))

        def make_app_name(dep):
            long_name = versioning.VersionSpec(dep).name
            return long_name.split('.', 1)[-1].replace('.', '')

        if 'requires' in lib_settings:
            for dep in lib_settings.get_list('requires', 'install', ''):
                app_name = make_app_name(dep)
                package_stack.append((app_name, None, dep))
                dependancy = True
            for dep in lib_settings.get_list('requires', 'mount', ''):
                app_name = make_app_name(dep)
                package_stack.append((app_name, "/{}/".format(app_name), dep))

    return requirements


if __name__ == "__main__":
    from moya.console import Console
    rpc = jsonrpc.JSONRPC('https://packages.moyaproject.com/jsonrpc/', ssl_verify=False)
    req = gather_dependencies(rpc, 'moya.logins==0.1.1-beta', Console())
    print(req)



from __future__ import unicode_literals
from ..compat import implements_to_string


@implements_to_string
class DataElement(object):
    def __init__(self, element, context):
        self.libid = element.libid
        self.ns = element.xmlns
        self.tag = element._tag_name
        self.data = element.get_all_data_parameters(context).copy()
        self.children = [DataElement(child, context) for child in element.children()]

    def __str__(self):
        return "<data {}>" .format(self.libid)

    def __repr__(self):
        return "<data {}>" .format(self.libid)



from __future__ import print_function
from __future__ import unicode_literals

import io

from . import iniparse
from . import errors

import logging
from logging import handlers
MemoryHandler = handlers.MemoryHandler

import fs.path
from fs.errors import FSError

from os.path import abspath, join, dirname, normpath


def _resolve(name):
    """Resolve a dotted name to a global object."""
    name = name.split('.')
    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used = used + '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)
    return found

_logging_level_names = {0: 'NOTSET',
                        10: 'DEBUG',
                        20: 'INFO',
                        30: 'WARNING',
                        40: 'ERROR',
                        50: 'CRITICAL',
                        'NOTSET': 0,
                        'DEBUG': 10,
                        'INFO': 20,
                        'WARN': 30,
                        'WARNING': 30,
                        'ERROR': 40,
                        'CRITICAL': 50}


def init_logging_fs(logging_fs, path, disable_existing_loggers=True):

    ini_path = path
    ini_stack = []
    while 1:
        try:
            with logging_fs.open(path, 'rt') as ini_file:
                s = iniparse.parse(ini_file)
        except FSError:
            raise errors.LoggingSettingsError('unable to read logging settings file "{}"'.format(path))
        ini_stack.append(s)
        if "extends" in s['']:
            path = fs.path.join(fs.path.dirname(path), s['']['extends'])
        else:
            break
    return _init_logging(ini_path, ini_stack, disable_existing_loggers)


def init_logging(path, disable_existing_loggers=True):
    """Sane logging.ini"""

    ini_path = path
    ini_stack = []
    visited = set()
    while 1:
        path = abspath(normpath(path))
        if path in visited:
            raise errors.LoggingSettingsError('recursive extends in logging ini')
        try:
            with io.open(path, 'rt') as ini_file:
                s = iniparse.parse(ini_file)
            visited.add(path)
        except IOError:
            raise errors.LoggingSettingsError('unable to read logging settings file "{}"'.format(path))
        ini_stack.append(s)
        if "extends" in s['']:
            path = join(dirname(path), s['']['extends'])
        else:
            break
    return _init_logging(ini_path, ini_stack, disable_existing_loggers)


def _init_logging(path, ini_stack, disable_existing_loggers=True):
    ini_path = path
    ini_stack = ini_stack[::-1]
    ini = ini_stack[0]
    for extend_ini in ini_stack[1:]:
        for section_name, section in extend_ini.items():
            if section_name in ini:
                ini[section_name].update(section)
            else:
                ini[section_name] = section

    def get(section_name, key, default=Ellipsis):
        try:
            value = ini[section_name][key]
        except KeyError:
            if default is Ellipsis:
                raise errors.LoggingSettingsError('unable to initialize logging (required key [{}]/{} was not found in "{}")'.format(section_name, key, ini_path))
            return default
        return value

    def getint(section_name, key, default=Ellipsis):
        value = get(section_name, key, default)
        if not value.isdigit():
            raise errors.LoggingSettingsError('unable to initialize logging (setting [{}]/{} should be an integer in "{path}")'.format(section_name, key, ini_path))
        return int(value)

    def getbool(section_name, key, default=Ellipsis):
        value = get(section_name, key, default).strip().lower()
        if value in ('yes', 'true'):
            return True
        if value in ('no', 'false'):
            return False
        raise errors.LoggingSettingsError('unable to initialize logging (section [{}]/{} is not a valid boolean in "{path}"'.format(section_name, key, ini_path))

    logging._acquireLock()
    try:
        _handlers = {}
        _formatters = {}
        _loggers = {}
        for section_name in ini:
            if not section_name:
                continue
            what, _, name = section_name.partition(':')
            if what == 'handler':
                _handlers[name] = section_name
            elif what == 'formatter':
                _formatters[name] = section_name
            elif what == 'logger':
                _loggers[name] = section_name
            else:
                raise errors.LoggingSettingsError('unable to initialize logging (section [{}] is not valid in "{}")'.format(section_name, ini_path))

        formatters = {}
        for formatter_name, section_name in _formatters.items():
            opts = ini[section_name]
            if "format" in opts:
                fs = get(section_name, "format")
            else:
                fs = None
            if "datefmt" in opts:
                dfs = get(section_name, "datefmt")
            else:
                dfs = None
            c = logging.Formatter
            if "class" in opts:
                class_name = get(section_name, "class")
                if class_name:
                    c = _resolve(class_name)
            f = c(fs, dfs)
            formatters[formatter_name] = f

        handlers = {}
        handlers['null'] = logging.NullHandler()
        fixups = []
        for handler_name, section_name in _handlers.items():
            klass = get(section_name, 'class')
            opts = ini[section_name]
            if "formatter" in opts:
                fmt = get(section_name, "formatter")
            else:
                fmt = ""
            try:
                kass = eval(klass, vars(logging))
            except (AttributeError, NameError):
                kass = _resolve(klass)
            args = get(section_name, "args")
            args = eval(args, vars(logging))
            try:
                h = kass(*args)
            except Exception as e:
                raise errors.LoggingSettingsError("error constructing logger '{}' with args {!r} ({})".format(kass, args, e))
            if "level" in opts:
                level = get(section_name, "level")
                h.setLevel(_logging_level_names[level])
            if len(fmt):
                h.setFormatter(formatters[fmt])
            if issubclass(kass, MemoryHandler):
                if "target" in opts:
                    target = get(section_name, "target")
                else:
                    target = ""
                if len(target):
                    fixups.append((h, target))
            handlers[handler_name] = h

        for h, t in fixups:
            h.setTarget(handlers[t])

        if 'root' not in _loggers:
            raise errors.LoggingSettingsError('unable to initialize logging (section [logger:root] is missing from "{}")'.format(ini_path))

        llist = list(_loggers.keys())
        llist.remove('root')

        root = logging.root
        log = root
        section_name = "logger:root"
        opts = ini[section_name]
        if "level" in opts:
            level = get(section_name, "level")
            log.setLevel(_logging_level_names[level])
        for h in root.handlers[:]:
            root.removeHandler(h)
        if 'handlers' in opts:
            hlist = [h.strip() for h in get(section_name, "handlers", 'null').split(',')]
            for hand in hlist:
                try:
                    log.addHandler(handlers[hand])
                except KeyError:
                    raise errors.LoggingSettingsError('''unable to initialize logging (handler '{}' not found in "{}")'''.format(hand, ini_path))

        existing = list(root.manager.loggerDict.keys())
        existing.sort()

        child_loggers = []
        for log in llist:
            section_name = "logger:{}".format(log)
            qn = get(section_name, "qualname", log)
            opts = ini[section_name]
            if "propagate" in opts:
                propagate = 1 if getbool(section_name, "propagate") else 0
            else:
                propagate = 1
            logger = logging.getLogger(qn)
            if qn in existing:
                i = existing.index(qn) + 1
                prefixed = qn + "."
                pflen = len(prefixed)
                num_existing = len(existing)
                while i < num_existing:
                    if existing[i][:pflen] == prefixed:
                        child_loggers.append(existing[i])
                    i += 1
                existing.remove(qn)
            if "level" in opts:
                level = get(section_name, "level", "NOTSET")
                if level not in _logging_level_names:
                    raise errors.LoggingSettingsError("unknown logging level '{}' in  [{}]".format(level, section_name))
                logger.setLevel(_logging_level_names[level])
            for h in logger.handlers[:]:
                logger.removeHandler(h)
            logger.propagate = propagate
            logger.disabled = 0

            if 'handlers' in opts:
                hlist = [h.strip() for h in get(section_name, "handlers", 'null').split(',')]
                for hand in hlist:
                    try:
                        logger.addHandler(handlers[hand])
                    except KeyError:
                        raise errors.LoggingSettingsError('''unable to initialize logging (handler '{}' not found in "{}")'''.format(hand, ini_path))

        for log in existing:
            logger = root.manager.loggerDict[log]
            if log in child_loggers:
                logger.level = logging.NOTSET
                logger.handlers = []
                logger.propagate = 1
            else:
                logger.disabled = disable_existing_loggers

    finally:
        logging._releaseLock()



from ...command import SubCommand
from ...tools import get_moya_dir
from ... import build
from ...console import Console

from fs.path import join
from fs.opener import fsopendir
from fs.tempfs import TempFS
from fs.osfs import OSFS

import sys
import os.path


class Doc(SubCommand):
    """Moya documentation"""
    help = """automatically generate Moya documentation"""

    def add_arguments(self, parser):

        parser.add_argument('--html', dest="html", action="store_true", default=False,
                            help="output should be HTML")

        subparsers = parser.add_subparsers(title="sub-commands",
                                           dest="action",
                                           help="sub command")

        extract_parser = subparsers.add_parser('extract',
                                               help="extract docs",
                                               description="Extract doc information")

        build_parser = subparsers.add_parser('build',
                                             help="build docs",
                                             description="Build extracted docs")

        view_parser = subparsers.add_parser('view',
                                            help="view docs",
                                            description="Extract and build docs, and launch the browser")

        def build_args(parser):
            # parser.add_argument('-b', '--lib', dest="lib", metavar="LONG.NAME", default=None,
            #                     help="library to generate docs for")
            parser.add_argument('-t', '--theme', dest='theme', metavar="PATH", default=None,
                                help="path to theme files (templates)")
            parser.add_argument('-s', '--source', dest="source", metavar="SOURCE", default=None,
                                help="path to extracted docs")
            parser.add_argument(dest="location", default='.', metavar="PATH",
                                help="location of library (directory containing lib.ini) or a python import if preceded by 'py:', e.g. py:moya.libs.auth")


        build_args(build_parser)
        build_args(view_parser)


        build_parser.add_argument('-o', '--output', dest="output", metavar="PATH", default=None,
                                  help="path for documentation output, defaults to ./documentation in project root")

        extract_parser.add_argument(dest="location", default=None, metavar="PATH",
                                    help="location of library (directory containing lib.ini) or a python import if preceded by 'py:', e.g. py:moya.libs.auth")

        # extract_parser.add_argument('-n', '--xmlns', dest="namespaces", metavar="XML NAMESPACE", action="append",
        #                             help="Namespace to generate docs for")
        extract_parser.add_argument('-e', '--extract', dest="extract", metavar="PATH", default=None,
                                    help="path to save raw documentation information")

        return parser

    def get_fs(self, path):
        if path is None:
            path = join(get_moya_dir(), './documentation')
        fs = fsopendir(path, create_dir=True)
        fs.dir_mode = int('777', 8)
        return fs

    def run(self):
        args = self.args

        _stdout = sys.stdout
        if args.html:
            self.console = Console(html=True)
            sys.stdout = self.console.make_file_interface()
            import moya
            moya.pilot.console = self.console

        try:
            action = args.action.lower()
            if action == 'extract':
                archive, lib = build.build_lib(args.location, ignore_errors=True)
                print("Extracting docs from {}...".format(lib.long_name))
                self.extract(archive, lib.long_name)

            elif action == 'build':
                if args.source is not None:
                    self.console.text("Building docs from {}...".format(args.source))
                    extract_fs = fsopendir(args.source)
                else:
                    archive, lib = build.build_lib(args.location, ignore_errors=True)
                    self.console.text("Building docs for {}...".format(lib.long_name))
                    extract_fs = self.extract(archive, lib.long_name)
                return self.build(extract_fs)

            elif action == 'view':
                if args.source is not None:
                    self.console.text("Building docs from {}...".format(args.source))
                    extract_fs = fsopendir(args.source)
                else:
                    archive, lib = build.build_lib(args.location, ignore_errors=True)
                    self.console.text("Building docs for {}...".format(lib.long_name))
                    extract_fs = self.extract(archive, lib.long_name)
                return self.view(archive, extract_fs)

            else:
                sys.error.write('action should be EXTRACT or BUILD\n')
                return -1
        finally:
            if args.html:
                _stdout.write(self.console.get_text() + '\n')

    def extract(self, archive, lib_name):
        args = self.args

        from ...docgen.extracter import Extracter

        if getattr(args, 'extract', None) is None:
            extract_fs = TempFS('moyadoc-{}'.format(lib_name))
        else:
            extract_fs = self.get_fs(join(args.extract, lib_name))
        extracter = Extracter(archive, extract_fs)
        extracter.extract_lib(lib_name)
        return extract_fs

    def build(self, source_fs):
        args = self.args
        output_fs = self.get_fs(args.output)

        out_path = output_fs.desc('/')

        # if not output_fs.isdirempty('/'):
        #     if raw_input('{} is not empty. Overwrite? (Y/N) '.format(out_path)).lower() not in ('y', 'yes'):
        #         sys.stdout.write('aborted\n')
        #         return -1

        if args.theme is None:
            from ... import docgen
            theme_path = os.path.join(os.path.dirname(docgen.__file__), 'themes/default')
        else:
            theme_path = args.theme
        theme_fs = self.get_fs(theme_path)

        from ...docgen.builder import Builder
        builder = Builder(source_fs, output_fs, theme_fs)
        builder.build()

    def view(self, archive, source_fs):
        args = self.args
        import tempfile
        docs_output_path = os.path.join(tempfile.tempdir, '__moyadocs__')
        output_fs = OSFS(docs_output_path, create=True)
        out_path = output_fs.desc('/')

        if args.theme is None:
            from ... import docgen
            theme_path = os.path.join(os.path.dirname(docgen.__file__), 'themes/default')
        else:
            theme_path = args.theme
        theme_fs = self.get_fs(theme_path)

        from ...docgen.builder import Builder
        builder = Builder(source_fs, output_fs, theme_fs)
        index = builder.build()
        import webbrowser
        webbrowser.open(index)



from __future__ import unicode_literals
from ..elements import Attribute
from ..elements.elementbase import LogicElement
from ..context.expressiontime import to_seconds


class CookieJar(dict):
    """A container for cookies"""

    def __init__(self, *args, **kwargs):
        self.deleted_cookies = set()
        super(CookieJar, self).__init__(*args, **kwargs)

    def __moyaconsole__(self, console):
        console.text("These are pending cookies that Moya will send with the response")
        from ..console import Cell
        table = [[Cell('name'),
                  Cell('value'),
                  Cell('path'),
                  Cell('domain'),
                  Cell('secure?')]]
        for cookie in self.itervalues():
            if isinstance(cookie, Cookie):
                table += [cookie.name,
                          cookie.value,
                          cookie.path,
                          cookie.domain,
                          'Y' if cookie.secure else 'N'
                          ]
        console.table(table)

    def __delitem__(self, key):
        self.deleted_cookies.add(key)
        super(self, '__delitem__')(key)


class Cookie(object):
    """Stores incoming cookies"""
    def __init__(self, name, value, max_age, path, domain, secure, httponly, comment, overwrite):
        self.name = name
        self.value = value
        self.max_age = max_age
        self.path = path
        self.domain = domain
        self.secure = secure
        self.httponly = httponly
        self.comment = comment
        self.overwrite = overwrite

    def __repr__(self):
        return "<cookie {} {!r}>".format(self.name, self.value)

    def set(self, response):
        response.set_cookie(self.name,
                            self.value,
                            max_age=self.max_age,
                            path=self.path,
                            domain=self.domain,
                            secure=self.secure,
                            httponly=self.httponly,
                            comment=self.comment,
                            overwrite=self.overwrite)


class SetCookie(LogicElement):
    """Set a new cookie."""

    class Help:
        synopsis = "set a cookie"
        example = """
            <set-cookie name="session" value="${session.key}" overwrite="yes" />
        """

    name = Attribute("Cookie name")
    value = Attribute("Value", required=True)
    maxage = Attribute("Max age of cookie (in seconds or as timespan)",
                       metavar="AGE",
                       type="timespan",
                       required=False,
                       example="60m")
    path = Attribute("Path", required=False, default="/", metavar="PATH")
    domain = Attribute("Domain", required=False, default=None, metavar="DOMAIN")
    secure = Attribute("Secure", type="boolean", default=False)
    httponly = Attribute("HTTP Only?", type="boolean", required=False, default=False)
    comment = Attribute("Comment", required=False, default=None)
    overwrite = Attribute("Overwrite?", required=False, default=False)

    def logic(self, context):
        params = self.get_parameters(context)
        cookie = Cookie(params.name,
                        params.value,
                        to_seconds(params.maxage),
                        params.path,
                        params.domain,
                        params.secure,
                        params.httponly,
                        params.comment,
                        params.overwrite)
        context.root['cookiejar'][params.name] = cookie


class DeleteCookie(LogicElement):
    """Delete a previously set cookie."""

    class Help:
        synopsis = "delete a cookie"
        example = """
            <delete-cookie name="authsession" />
        """

    name = Attribute("Cookie name")

    def logic(self, context):
        try:
            context.root['cookiejar'][self.name(context)]
        except KeyError:
            pass



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

from ...command import SubCommand
from ...wsgi import WSGIApplication
from ...console import Cell
from ...compat import text_type, raw_input

from fs.opener import fsopendir
from fs.errors import FSError
from fs.multifs import MultiFS
from fs.mountfs import MountFS

import os


def _ls(console, file_paths, dir_paths, format_long=False):
    """Cannibalized from pyfileystem"""

    dirs = frozenset(dir_paths)
    paths = sorted(file_paths + dir_paths, key=lambda p: p.lower())

    def columnize(paths, num_columns):
        col_height = (len(paths) + num_columns - 1) / num_columns
        columns = [[] for _ in range(num_columns)]
        col_no = 0
        col_pos = 0
        for path in paths:
            columns[col_no].append(path)
            col_pos += 1
            if col_pos >= col_height:
                col_no += 1
                col_pos = 0

        padded_columns = []

        def wrap(path):
            return (path in dirs, path.ljust(max_width))

        for column in columns:
            if column:
                max_width = max([len(path) for path in column])
            else:
                max_width = 1
            max_width = min(max_width, terminal_width)
            padded_columns.append([wrap(path) for path in column])

        return padded_columns

    def condense_columns(columns):
        max_column_height = max([len(col) for col in columns])
        lines = [[] for _ in range(max_column_height)]
        for column in columns:
            for line, (isdir, path) in zip(lines, column):
                line.append((isdir, path))
        for line in lines:
            for i, (isdir, path) in enumerate(line):
                if isdir:
                    console(path, bold=True, fg="blue")
                else:
                    console(path)
                if i < len(line) - 1:
                    console('  ')
            console.nl()

    if format_long:
        for path in paths:
            if path in dirs:
                console(path, bold=True, fg="blue")
            else:
                console(path)
            console.nl()

    else:
        terminal_width = console.width
        path_widths = [len(path) for path in paths]
        smallest_paths = min(path_widths)
        num_paths = len(paths)

        num_cols = min(terminal_width // (smallest_paths + 2), num_paths)
        while num_cols:
            col_height = (num_paths + num_cols - 1) // num_cols
            line_width = 0
            for col_no in range(num_cols):
                try:
                    col_width = max(path_widths[col_no * col_height: (col_no + 1) * col_height])
                except ValueError:
                    continue
                line_width += col_width
                if line_width > terminal_width:
                    break
                line_width += 2
            else:
                if line_width - 1 <= terminal_width:
                    break
            num_cols -= 1
        num_cols = max(1, num_cols)
        columns = columnize(paths, num_cols)
        condense_columns(columns)


class FS(SubCommand):
    """Manage project filesystems"""
    help = "manage project fsfilesystems"

    def add_arguments(self, parser):
        parser.add_argument(dest="fs", nargs="?", default=None, metavar="FILESYSTEM",
                            help="filesystem name")
        parser.add_argument("-l", "--location", dest="location", default=None, metavar="PATH",
                            help="location of the Moya server code")
        parser.add_argument("-i", "--ini", dest="settings", default=None, metavar="SETTINGSPATH",
                            help="path to project settings")
        parser.add_argument("--server", dest="server", default='main', metavar="SERVERREF",
                            help="server element to use")
        parser.add_argument('--ls', dest="listdir", default=None, metavar="PATH",
                            help="list files / directories")
        parser.add_argument("--tree", dest="tree", nargs='?', default=None, const='/',
                            help="display a tree view of the filesystem")
        parser.add_argument("--cat", dest="cat", default=None, metavar="PATH",
                            help="Cat a file to the console")
        parser.add_argument("--syspath", dest="syspath", default=None, metavar="PATH",
                            help="display the system path of a file")
        parser.add_argument("--open", dest="open", default=None, metavar="PATH",
                            help="open a file")
        parser.add_argument("--copy", dest="copy", metavar="PATH", nargs='+',
                            help="copy contents of filesystem to PATH")
        parser.add_argument("-f", "--force", dest="force", action="store_true", default=False,
                            help="force overwrite of destination even if it is not empty (with --copy)")
        return parser

    def run(self):
        args = self.args
        application = WSGIApplication(self.location,
                                      self.get_settings(),
                                      args.server,
                                      disable_autoreload=True)
        archive = application.archive

        filesystems = archive.filesystems

        fs = None
        if args.fs:
            try:
                fs = filesystems[args.fs]
            except KeyError:
                self.console.error("No filesystem called '%s'" % args.fs)
                return -1

        if args.tree is not None:
            if fs is None:
                self.console.error("Filesystem required")
                return -1
            with fs.opendir(args.tree) as tree_fs:
                tree_fs.tree()
            return

        if args.listdir:
            if fs is None:
                self.console.error("Filesystem required")
                return -1

            dir_fs = fs.opendir(args.listdir)
            file_paths = dir_fs.listdir(files_only=True)
            dir_paths = dir_fs.listdir(dirs_only=True)
            _ls(self.console, file_paths, dir_paths)

        elif args.cat:
            if fs is None:
                self.console.error("Filesystem required")
                return -1
            contents = fs.getcontents(args.cat)
            self.console.cat(contents, args.cat)

        elif args.open:
            if fs is None:
                self.console.error("Filesystem required")
                return -1

            filepath = fs.getsyspath(args.open, allow_none=True)
            if filepath is None:
                self.console.error("No system path for '%s' in filesystem '%s'" % (args.open, args.fs))
                return -1

            import subprocess
            if os.name == 'mac':
                subprocess.call(('open', filepath))
            elif os.name == 'nt':
                subprocess.call(('start', filepath), shell=True)
            elif os.name == 'posix':
                subprocess.call(('xdg-open', filepath))
            else:
                self.console.error("Don't know how to open files on this platform (%s)" % os.name)

        elif args.syspath:
            if fs is None:
                self.console.error("Filesystem required (use -cat FILESYSTEM)")
                return -1
            if not fs.exists(args.syspath):
                self.console.error("No file called '%s' found in filesystem '%s'" % (args.syspath, args.fs))
                return -1
            syspath = fs.getsyspath(args.syspath, allow_none=True)
            if syspath is None:
                self.console.error("No system path for '%s' in filesystem '%s'" % (args.syspath, args.fs))
            else:
                self.console(syspath).nl()

        elif args.copy:
            if len(args.copy) == 1:
                src = '/'
                dst = args.copy[0]
            elif len(args.copy) == 2:
                src, dst = args.copy
            else:
                self.console.error("--copy requires 1 or 2 arguments")
                return -1

            if fs.isdir(src):
                src_fs = fs.opendir(src)
                dst_fs = fsopendir(dst, create_dir=True)

                if not args.force and not dst_fs.isdirempty('/'):
                    response = raw_input("'%s' is not empty. Copying may overwrite directory contents. Continue? " % dst)
                    if response.lower() not in ('y', 'yes'):
                        return 0

                from fs.utils import copydir
                copydir(src_fs, dst_fs)
            else:
                with fs.open(src, 'rb') as read_f:
                    if os.path.isdir(dst):
                        dst = os.path.join(dst, os.path.basename(src))
                    with open(dst, 'wb') as write_f:
                        while 1:
                            chunk = read_f.read(16384)
                            if not chunk:
                                break
                            write_f.write(chunk)

        else:
            table = [[Cell("Name", bold=True),
                      Cell("Type", bold=True),
                      Cell("Location", bold=True)]]

            if fs is None:
                list_filesystems = filesystems.items()
            else:
                list_filesystems = [(args.fs, fs)]

            for name, fs in sorted(list_filesystems):

                if isinstance(fs, MultiFS):
                    location = '\n'.join(mount_fs.desc('/') for mount_fs in fs.fs_sequence)
                    fg = "yellow"
                elif isinstance(fs, MountFS):
                    mount_desc = []
                    for path, dirmount in fs.mount_tree.items():
                        mount_desc.append('%s->%s' % (path, dirmount.fs.desc('/')))
                    location = '\n'.join(mount_desc)
                    fg = "magenta"
                else:
                    syspath = fs.getsyspath('/', allow_none=True)
                    if syspath is not None:
                        location = syspath
                        fg = "green"
                    else:
                        try:
                            location = fs.desc('/')
                        except FSError as e:
                            location = text_type(e)
                            fg = "red"
                        else:
                            fg = "blue"
                table.append([Cell(name),
                             Cell(type(fs).__name__),
                             Cell(location, bold=True, fg=fg)
                              ])
            self.console.table(table, header=True)



from __future__ import unicode_literals


class BoundElement(object):

    __slots__ = ('app', 'element')

    def __init__(self, app, element):
        self.app = app
        self.element = element

    @classmethod
    def from_tuple(cls, app_element):
        app, element = app_element
        return cls(app, element)

    def __repr__(self):
        return "<{1} in app {0}>".format(self.app, self.element)

    def __iter__(self):
        return iter(('app', 'element'))

    def __getitem__(self, key):
        if key == 'app':
            return self.app
        elif key == 'element':
            return self.element
        raise KeyError(key)

    def __moyamodel__(self):
        return self.app, self.element

    def keys(self):
        return ('app', 'element')

    def iterkeys(self):
        return iter(self.keys())

    def values(self):
        return (self.app, self.element)

    def itervalues(self):
        return iter(self.values())

    def items(self):
        return [('app', self.app), ('element', self.element)]

    def iteritems(self):
        return self.items()



from __future__ import unicode_literals
from __future__ import print_function

from ..compat import string_types
from .errors import MissingTemplateError, BadTemplateError
from ..context.expressiontime import datetime_to_epoch
from .moyatemplates import Template
from ..cache.dictcache import DictCache

from fs.path import abspath, normpath
from fs.opener import fsopendir
from fs.errors import ResourceNotFoundError, NoSysPathError

import weakref


# Update for backwards incompatible changes, so we don't get old cached templates
TEMPLATE_VERSION = 8


class Environment(object):
    name = "moya"

    def __init__(self, template_fs, archive=None, cache=None):
        if isinstance(template_fs, string_types):
            template_fs = fsopendir(template_fs)
        self.template_fs = template_fs
        if archive is not None:
            self._archive = weakref.ref(archive)
            self.cache = cache or archive.get_cache('templates')
        else:
            self._archive = lambda: None
            if cache is None:
                cache = DictCache("templates", "")
            self.cache = cache
        self.templates = {}
        self._caches = {}

    @classmethod
    def make_default(self):
        return Environment(fsopendir('mem://'))

    @property
    def archive(self):
        return self._archive()

    def check_template(self, template_path):
        """Check if a template exists"""
        template_path = abspath(normpath(template_path))
        if template_path in self.templates:
            return
        if not self.template_fs.exists(template_path):
            raise MissingTemplateError(template_path)

    def get_template(self, template_path, parse=True):
        template_path = abspath(normpath(template_path))
        template = self.templates.get(template_path, None)
        if template is not None:
            return template
        if self.cache.enabled:
            try:
                info = self.template_fs.getinfokeys(template_path, 'modified_time', 'st_mtime')
            except ResourceNotFoundError:
                raise MissingTemplateError(template_path)
            modified_time = info.get('st_mtime', None) or datetime_to_epoch(info.get('modified_time', 0))
            cache_name = "%s$%s@%s" % (TEMPLATE_VERSION, template_path, modified_time)
            cached_template = self.cache.get(cache_name, None)
            if cached_template is not None:
                template = Template.load(cached_template)
                self.templates[template_path] = template
                return template

        try:
            source = self.template_fs.getcontents(template_path, 'rt', encoding='utf-8')
        except ResourceNotFoundError:
            raise MissingTemplateError(template_path)
        except Exception as e:
            raise BadTemplateError(template_path)

        try:
            display_path = self.template_fs.getsyspath(template_path)
        except NoSysPathError:
            display_path = template_path
        if self.archive is None:
            lib = None
        else:
            lib = self.archive.get_template_lib(template_path)

        template = Template(source, display_path, raw_path=template_path, lib=lib)
        template.parse(self)

        if self.cache.enabled:
            self.cache.set(cache_name, template.dump(self))

        self.templates[template_path] = template
        return template

    def compile_template(self, source, path):
        template = Template(source, path)
        return template

    def get_cache(self, cache_name):
        if self.archive is not None:
            return self.archive.get_cache(cache_name)
        if cache_name not in self._caches:
            self._caches[cache_name] = DictCache(cache_name, "")
        return self._caches[cache_name]




from .context.context import Context
from .compat import text_type

import json
import hashlib
import logging
log = logging.getLogger('moya.runtine')


DEFAULT = """
{
    "_moya":
    {
        "note": "internal default"
    },
    "colors":
    {
        "text":
        {
            "fg": "black",
            "bg": "white"
        },
        "highlight":
        {
            "fg": "black",
            "bg": "wheat"
        },
        "selected":
        {
            "fg": "white",
            "bg": "blue"
        },
        "border":
        {
            "normal": "inherit",
            "focused": "inherit"
        }
    }

}
"""


class Theme(object):

    _cache = {}

    @classmethod
    def loader(cls, fs):
        """Automatically add theme to context"""
        def load(context):
            name = context.get('.sys.site.theme', 'default')

            path = "{}.json".format(name)
            try:
                theme = cls.read(fs, path, context=context)
            except Exception as e:
                log.warning("unable to read theme file '%s' (%s)", path, text_type(e))

                if name != 'default':
                    return load('default')

                log.error("unable to load 'default' theme")
                theme = None

            return theme

        return load

    @classmethod
    def dummy_loader(cls, context):
        """Called when theme is not enabled"""
        log.warning('theme is not set -- add a theme value to your site settings')
        theme_json = DEFAULT
        hasher = hashlib.md5()
        hasher.update(theme_json)
        theme_hash = hasher.hexdigest()

        theme_data = json.loads(theme_json)

        theme_data['_moya'] = {
            "path": None,
            "hash": theme_hash
        }
        return theme_data

    @classmethod
    def read(cls, fs, path, context=None):

        if context is None:
            context = Context()

        hasher = hashlib.md5()
        with fs.open(path, 'rb') as f:
            theme_json = f.read()
            hasher.update(theme_json)
            theme_hash = hasher.hexdigest()

            theme_data = json.loads(theme_json)

            theme_data['_moya'] = {
                "path": path,
                "hash": theme_hash
            }

        return theme_data
